@@ -1,7 +1,6 @@
 #!/bin/bash
-# Copyright 1999-2009 Gentoo Foundation
+# Copyright 1999-2010 Gentoo Foundation
 # Distributed under the terms of the GNU General Public License v2
-# $Id: misc-functions.sh 14950 2009-12-06 11:03:54Z zmedico $
 #
 # Miscellaneous shell functions that make use of the ebuild env but don't need
 # to be included directly in ebuild.sh.
@@ -41,17 +40,134 @@ install_symlink_html_docs() {
 	fi
 }
 
+# replacement for "readlink -f" or "realpath"
+canonicalize() {
+	local f=$1 b n=10 wd=$(pwd)
+	while (( n-- > 0 )); do
+		while [[ ${f: -1} = / && ${#f} -gt 1 ]]; do
+			f=${f%/}
+		done
+		b=${f##*/}
+		cd "${f%"${b}"}" 2>/dev/null || break
+		if [[ ! -L ${b} ]]; then
+			f=$(pwd -P)
+			echo "${f%/}/${b}"
+			cd "${wd}"
+			return 0
+		fi
+		f=$(readlink "${b}")
+	done
+	cd "${wd}"
+	return 1
+}
+
+prepcompress() {
+	local -a include exclude incl_d incl_f
+	local f g i real_f real_d
+
+	# Canonicalize path names and check for their existence.
+	real_d=$(canonicalize "${D}")
+	for (( i = 0; i < ${#PORTAGE_DOCOMPRESS[@]}; i++ )); do
+		real_f=$(canonicalize "${D}${PORTAGE_DOCOMPRESS[i]}")
+		f=${real_f#"${real_d}"}
+		if [[ ${real_f} != "${f}" ]] && [[ -d ${real_f} || -f ${real_f} ]]
+		then
+			include[${#include[@]}]=${f:-/}
+		elif [[ ${i} -ge 3 ]]; then
+			ewarn "prepcompress:" \
+				"ignoring nonexistent path '${PORTAGE_DOCOMPRESS[i]}'"
+		fi
+	done
+	for (( i = 0; i < ${#PORTAGE_DOCOMPRESS_SKIP[@]}; i++ )); do
+		real_f=$(canonicalize "${D}${PORTAGE_DOCOMPRESS_SKIP[i]}")
+		f=${real_f#"${real_d}"}
+		if [[ ${real_f} != "${f}" ]] && [[ -d ${real_f} || -f ${real_f} ]]
+		then
+			exclude[${#exclude[@]}]=${f:-/}
+		elif [[ ${i} -ge 1 ]]; then
+			ewarn "prepcompress:" \
+				"ignoring nonexistent path '${PORTAGE_DOCOMPRESS_SKIP[i]}'"
+		fi
+	done
+
+	# Remove redundant entries from lists.
+	# For the include list, remove any entries that are:
+	# a) contained in a directory in the include or exclude lists, or
+	# b) identical with an entry in the exclude list.
+	for (( i = ${#include[@]} - 1; i >= 0; i-- )); do
+		f=${include[i]}
+		for g in "${include[@]}"; do
+			if [[ ${f} == "${g%/}"/* ]]; then
+				unset include[i]
+				continue 2
+			fi
+		done
+		for g in "${exclude[@]}"; do
+			if [[ ${f} = ${g} || ${f} == "${g%/}"/* ]]; then
+				unset include[i]
+				continue 2
+			fi
+		done
+	done
+	# For the exclude list, remove any entries that are:
+	# a) contained in a directory in the exclude list, or
+	# b) _not_ contained in a directory in the include list.
+	for (( i = ${#exclude[@]} - 1; i >= 0; i-- )); do
+		f=${exclude[i]}
+		for g in "${exclude[@]}"; do
+			if [[ ${f} == "${g%/}"/* ]]; then
+				unset exclude[i]
+				continue 2
+			fi
+		done
+		for g in "${include[@]}"; do
+			[[ ${f} == "${g%/}"/* ]] && continue 2
+		done
+		unset exclude[i]
+	done
+
+	# Split the include list into directories and files
+	for f in "${include[@]}"; do
+		if [[ -d ${D}${f} ]]; then
+			incl_d[${#incl_d[@]}]=${f}
+		else
+			incl_f[${#incl_f[@]}]=${f}
+		fi
+	done
+
+	# Queue up for compression.
+	# ecompress{,dir} doesn't like to be called with empty argument lists.
+	[[ ${#incl_d[@]} -gt 0 ]] && ecompressdir --queue "${incl_d[@]}"
+	[[ ${#incl_f[@]} -gt 0 ]] && ecompress --queue "${incl_f[@]/#/${D}}"
+	[[ ${#exclude[@]} -gt 0 ]] && ecompressdir --ignore "${exclude[@]}"
+	return 0
+}
+
 install_qa_check() {
+	local f x
+
 	cd "${D}" || die "cd failed"
 
 	export STRIP_MASK
 	prepall
+	hasq "${EAPI}" 0 1 2 3 || prepcompress
 	ecompressdir --dequeue
 	ecompress --dequeue
 
+	f=
+	for x in etc/app-defaults usr/man usr/info usr/X11R6 usr/doc usr/locale ; do
+		[[ -d $D/$x ]] && f+="  $x\n"
+	done
+
+	if [[ -n $f ]] ; then
+		eqawarn "QA Notice: This ebuild installs into the following deprecated directories:"
+		eqawarn
+		eqawarn "$f"
+	fi
+
 	# Now we look for all world writable files.
+	local i
 	for i in $(find "${D}/" -type f -perm -2); do
-		vecho -ne '\a'
 		vecho "QA Security Notice:"
 		vecho "- ${i:${#D}:${#i}} will be a world writable file."
 		vecho "- This may or may not be a security problem, most of the time it is one."
@@ -61,7 +177,7 @@ install_qa_check() {
 
 	if type -P scanelf > /dev/null && ! hasq binchecks ${RESTRICT}; then
 		local qa_var insecure_rpath=0 tmp_quiet=${PORTAGE_QUIET}
-		local f x
+		local x
 
 		# display warnings when using stricter because we die afterwards
 		if has stricter ${FEATURES} ; then
@@ -79,12 +195,12 @@ install_qa_check() {
 			xargs -0 scanelf -qyRF '%r %p' | grep '$ORIGIN'
 		)
 		if [[ -n ${f}${x} ]] ; then
-			vecho -ne '\a\n'
+			vecho -ne '\n'
 			eqawarn "QA Notice: The following files contain insecure RUNPATHs"
 			eqawarn " Please file a bug about this at http://bugs.gentoo.org/"
 			eqawarn " with the maintaining herd of the package."
 			eqawarn "${f}${f:+${x:+\n}}${x}"
-			vecho -ne '\a\n'
+			vecho -ne '\n'
 			if [[ -n ${x} ]] || has stricter ${FEATURES} ; then
 				insecure_rpath=1
 			else
@@ -105,7 +221,7 @@ install_qa_check() {
 		f=$(scanelf -qyRF '%t %p' "${D}" | grep -v 'usr/lib/debug/')
 		if [[ -n ${f} ]] ; then
 			scanelf -qyRAF '%T %p' "${PORTAGE_BUILDDIR}"/ &> "${T}"/scanelf-textrel.log
-			vecho -ne '\a\n'
+			vecho -ne '\n'
 			eqawarn "QA Notice: The following files contain runtime text relocations"
 			eqawarn " Text relocations force the dynamic linker to perform extra"
 			eqawarn " work at startup, waste system resources, and may pose a security"
@@ -114,7 +230,7 @@ install_qa_check() {
 			eqawarn " For more information, see http://hardened.gentoo.org/pic-fix-guide.xml"
 			eqawarn " Please include the following list of files in your report:"
 			eqawarn "${f}"
-			vecho -ne '\a\n'
+			vecho -ne '\n'
 			die_msg="${die_msg} textrels,"
 			sleep 1
 		fi
@@ -150,7 +266,7 @@ install_qa_check() {
 		if [[ -n ${f} ]] ; then
 			# One more pass to help devs track down the source
 			scanelf -qyRAF '%e %p' "${PORTAGE_BUILDDIR}"/ &> "${T}"/scanelf-execstack.log
-			vecho -ne '\a\n'
+			vecho -ne '\n'
 			eqawarn "QA Notice: The following files contain writable and executable sections"
 			eqawarn " Files with such sections will not work properly (or at all!) on some"
 			eqawarn " architectures/operating systems.  A bug should be filed at"
@@ -160,13 +276,13 @@ install_qa_check() {
 			eqawarn " Note: Bugs should be filed for the respective maintainers"
 			eqawarn " of the package in question and not hardened@g.o."
 			eqawarn "${f}"
-			vecho -ne '\a\n'
+			vecho -ne '\n'
 			die_msg="${die_msg} execstacks"
 			sleep 1
 		fi
 
 		# Check for files built without respecting LDFLAGS
-		if [[ "${LDFLAGS}" == *--hash-style=gnu* ]] && [[ "${PN}" != *-bin ]] ; then
+		if [[ "${LDFLAGS}" == *,--hash-style=gnu* ]] && [[ "${PN}" != *-bin ]] ; then
 			qa_var="QA_DT_HASH_${ARCH/-/_}"
 			eval "[[ -n \${!qa_var} ]] && QA_DT_HASH=(\"\${${qa_var}[@]}\")"
 			f=$(scanelf -qyRF '%k %p' -k .hash "${D}" | sed -e "s:\.hash ::")
@@ -193,11 +309,11 @@ install_qa_check() {
 					-i "${T}"/scanelf-ignored-LDFLAGS.log
 				f=$(<"${T}"/scanelf-ignored-LDFLAGS.log)
 				if [[ -n ${f} ]] ; then
-					vecho -ne '\a\n'
+					vecho -ne '\n'
 					eqawarn "${BAD}QA Notice: Files built without respecting LDFLAGS have been detected${NORMAL}"
 					eqawarn " Please include the following list of files in your report:"
 					eqawarn "${f}"
-					vecho -ne '\a\n'
+					vecho -ne '\n'
 					sleep 1
 				else
 					rm -f "${T}"/scanelf-ignored-LDFLAGS.log
@@ -267,10 +383,10 @@ install_qa_check() {
 			sed -e "/^\$/d" -i "${T}"/scanelf-missing-SONAME.log
 			f=$(<"${T}"/scanelf-missing-SONAME.log)
 			if [[ -n ${f} ]] ; then
-				vecho -ne '\a\n'
+				vecho -ne '\n'
 				eqawarn "QA Notice: The following shared libraries lack a SONAME"
 				eqawarn "${f}"
-				vecho -ne '\a\n'
+				vecho -ne '\n'
 				sleep 1
 			else
 				rm -f "${T}"/scanelf-missing-SONAME.log
@@ -280,7 +396,7 @@ install_qa_check() {
 		# Check for shared libraries lacking NEEDED entries
 		qa_var="QA_DT_NEEDED_${ARCH/-/_}"
 		eval "[[ -n \${!qa_var} ]] && QA_DT_NEEDED=(\"\${${qa_var}[@]}\")"
-		f=$(scanelf -ByF '%n %p' "${D}"{,usr/}lib*.so* | gawk '$2 == "" { print }' | sed -e "s:^[[:space:]]${D}:/:")
+		f=$(scanelf -ByF '%n %p' "${D}"{,usr/}lib*/lib*.so* | gawk '$2 == "" { print }' | sed -e "s:^[[:space:]]${D}:/:")
 		if [[ -n ${f} ]] ; then
 			echo "${f}" > "${T}"/scanelf-missing-NEEDED.log
 			if [[ "${QA_STRICT_DT_NEEDED-unset}" == unset ]] ; then
@@ -301,10 +417,10 @@ install_qa_check() {
 			sed -e "/^\$/d" -i "${T}"/scanelf-missing-NEEDED.log
 			f=$(<"${T}"/scanelf-missing-NEEDED.log)
 			if [[ -n ${f} ]] ; then
-				vecho -ne '\a\n'
+				vecho -ne '\n'
 				eqawarn "QA Notice: The following shared libraries lack NEEDED entries"
 				eqawarn "${f}"
-				vecho -ne '\a\n'
+				vecho -ne '\n'
 				sleep 1
 			else
 				rm -f "${T}"/scanelf-missing-NEEDED.log
@@ -331,8 +447,21 @@ install_qa_check() {
 		unset INSTALLTOD
 	fi
 
+	# Sanity check syntax errors in init.d scripts
+	local d
+	for d in /etc/conf.d /etc/init.d ; do
+		[[ -d ${D}/${d} ]] || continue
+		for i in "${D}"/${d}/* ; do
+			[[ -L ${i} ]] && continue
+			# if empty conf.d/init.d dir exists (baselayout), then i will be "/etc/conf.d/*" and not exist
+			[[ ! -e ${i} ]] && continue
+			bash -n "${i}" || die "The init.d file has syntax errors: ${i}"
+		done
+	done
+
 	# this should help to ensure that all (most?) shared libraries are executable
 	# and that all libtool scripts / static libraries are not executable
+	local j
 	for i in "${D}"opt/*/lib{,32,64} \
 	         "${D}"lib{,32,64}       \
 	         "${D}"usr/lib{,32,64}   \
@@ -341,17 +470,7 @@ install_qa_check() {
 
 		for j in "${i}"/*.so.* "${i}"/*.so ; do
 			[[ ! -e ${j} ]] && continue
-			if [[ -L ${j} ]] ; then
-				linkdest=$(readlink "${j}")
-				if [[ ${linkdest} == /* ]] ; then
-					vecho -ne '\a\n'
-					eqawarn "QA Notice: Found an absolute symlink in a library directory:"
-					eqawarn "           ${j#${D}} -> ${linkdest}"
-					eqawarn "           It should be a relative symlink if in the same directory"
-					eqawarn "           or a linker script if it crosses the /usr boundary."
-				fi
-				continue
-			fi
+			[[ -L ${j} ]] && continue
 			[[ -x ${j} ]] && continue
 			vecho "making executable: ${j#${D}}"
 			chmod +x "${j}"
@@ -364,6 +483,19 @@ install_qa_check() {
 			vecho "removing executable bit: ${j#${D}}"
 			chmod -x "${j}"
 		done
+
+		for j in "${i}"/*.{a,dll,dylib,sl,so}.* "${i}"/*.{a,dll,dylib,sl,so} ; do
+			[[ ! -e ${j} ]] && continue
+			[[ ! -L ${j} ]] && continue
+			linkdest=$(readlink "${j}")
+			if [[ ${linkdest} == /* ]] ; then
+				vecho -ne '\n'
+				eqawarn "QA Notice: Found an absolute symlink in a library directory:"
+				eqawarn "           ${j#${D}} -> ${linkdest}"
+				eqawarn "           It should be a relative symlink if in the same directory"
+				eqawarn "           or a linker script if it crosses the /usr boundary."
+			fi
+		done
 	done
 
 	# When installing static libraries into /usr/lib and shared libraries into 
@@ -371,12 +503,13 @@ install_qa_check() {
 	# the static library, or gcc will utilize the static lib when linking :(.
 	# http://bugs.gentoo.org/4411
 	abort="no"
+	local a s
 	for a in "${D}"usr/lib*/*.a ; do
 		s=${a%.a}.so
 		if [[ ! -e ${s} ]] ; then
 			s=${s%usr/*}${s##*/usr/}
 			if [[ -e ${s} ]] ; then
-				vecho -ne '\a\n'
+				vecho -ne '\n'
 				eqawarn "QA Notice: Missing gen_usr_ldscript for ${s##*/}"
 	 			abort="yes"
 			fi
@@ -387,19 +520,19 @@ install_qa_check() {
 	# Make sure people don't store libtool files or static libs in /lib
 	f=$(ls "${D}"lib*/*.{a,la} 2>/dev/null)
 	if [[ -n ${f} ]] ; then
-		vecho -ne '\a\n'
+		vecho -ne '\n'
 		eqawarn "QA Notice: Excessive files found in the / partition"
 		eqawarn "${f}"
-		vecho -ne '\a\n'
+		vecho -ne '\n'
 		die "static archives (*.a) and libtool library files (*.la) do not belong in /"
 	fi
 
 	# Verify that the libtool files don't contain bogus $D entries.
-	local abort=no gentoo_bug=no
+	local abort=no gentoo_bug=no always_overflow=no
 	for a in "${D}"usr/lib*/*.la ; do
 		s=${a##*/}
 		if grep -qs "${D}" "${a}" ; then
-			vecho -ne '\a\n'
+			vecho -ne '\n'
 			eqawarn "QA Notice: ${s} appears to contain PORTAGE_TMPDIR paths"
 			abort="yes"
 		fi
@@ -423,25 +556,74 @@ install_qa_check() {
 			": warning: is used uninitialized in this function$" # we'll ignore "may" and "might"
 			": warning: comparisons like X<=Y<=Z do not have their mathematical meaning$"
 			": warning: null argument where non-null required "
+			": warning: array subscript is below array bounds$"
+			": warning: array subscript is above array bounds$"
+			": warning: attempt to free a non-heap object"
+			": warning: .* called with .*bigger.* than .* destination buffer$"
+			": warning: call to .* will always overflow destination buffer$"
+			": warning: assuming pointer wraparound does not occur when comparing "
+			": warning: hex escape sequence out of range$"
+			": warning: [^ ]*-hand operand of comma .*has no effect$"
+			": warning: converting to non-pointer type .* from NULL"
+			": warning: NULL used in arithmetic$"
+			": warning: passing NULL to non-pointer argument"
+			": warning: the address of [^ ]* will always evaluate as"
+			": warning: the address of [^ ]* will never be NULL"
+			": warning: too few arguments for format"
+			": warning: reference to local variable .* returned"
+			": warning: returning reference to temporary"
+			": warning: function returns address of local variable"
+			# this may be valid code :/
+			#": warning: multi-character character constant$"
+			# need to check these two ...
+			#": warning: assuming signed overflow does not occur when "
+			#": warning: comparison with string literal results in unspecified behav"
+			# yacc/lex likes to trigger this one
+			#": warning: extra tokens at end of .* directive"
+			# only gcc itself triggers this ?
+			#": warning: .*noreturn.* function does return"
+			# these throw false positives when 0 is used instead of NULL
+			#": warning: missing sentinel in function call"
+			#": warning: not enough variable arguments to fit a sentinel"
 		)
 		abort="no"
 		i=0
+		local grep_cmd=grep
+		[[ $PORTAGE_LOG_FILE = *.gz ]] && grep_cmd=zgrep
 		while [[ -n ${msgs[${i}]} ]] ; do
 			m=${msgs[$((i++))]}
 			# force C locale to work around slow unicode locales #160234
-			f=$(LC_ALL=C grep "${m}" "${PORTAGE_LOG_FILE}")
+			f=$(LC_ALL=C $grep_cmd "${m}" "${PORTAGE_LOG_FILE}")
 			if [[ -n ${f} ]] ; then
-				vecho -ne '\a\n'
-				eqawarn "QA Notice: Package has poor programming practices which may compile"
-				eqawarn "           fine but exhibit random runtime failures."
-				eqawarn "${f}"
-				vecho -ne '\a\n'
 				abort="yes"
+				# for now, don't make this fatal (see bug #337031)
+				#case "$m" in
+				#	": warning: call to .* will always overflow destination buffer$") always_overflow=yes ;;
+				#esac
+				if [[ $always_overflow = yes ]] ; then
+					eerror
+					eerror "QA Notice: Package has poor programming practices which may compile"
+					eerror "           fine but exhibit random runtime failures."
+					eerror
+					eerror "${f}"
+					eerror
+					eerror " Please file a bug about this at http://bugs.gentoo.org/"
+					eerror " with the maintaining herd of the package."
+					eerror
+				else
+					vecho -ne '\n'
+					eqawarn "QA Notice: Package has poor programming practices which may compile"
+					eqawarn "           fine but exhibit random runtime failures."
+					eqawarn "${f}"
+					vecho -ne '\n'
+				fi
 			fi
 		done
+		local cat_cmd=cat
+		[[ $PORTAGE_LOG_FILE = *.gz ]] && cat_cmd=zcat
 		[[ $reset_debug = 1 ]] && set -x
-		f=$(cat "${PORTAGE_LOG_FILE}" | \
-			EPYTHON= "$PORTAGE_BIN_PATH"/check-implicit-pointer-usage.py)
+		f=$($cat_cmd "${PORTAGE_LOG_FILE}" | \
+			"${PORTAGE_PYTHON:-/usr/bin/python}" "$PORTAGE_BIN_PATH"/check-implicit-pointer-usage.py || die "check-implicit-pointer-usage.py failed")
 		if [[ -n ${f} ]] ; then
 
 			# In the future this will be a forced "die". In preparation,
@@ -466,16 +648,16 @@ install_qa_check() {
 				eerror " with the maintaining herd of the package."
 				eerror
 			else
-				vecho -ne '\a\n'
+				vecho -ne '\n'
 				eqawarn "QA Notice: Package has poor programming practices which may compile"
 				eqawarn "           but will almost certainly crash on 64bit architectures."
 				eqawarn "${f}"
-				vecho -ne '\a\n'
+				vecho -ne '\n'
 			fi
 
 		fi
 		if [[ ${abort} == "yes" ]] ; then
-			if [[ ${gentoo_bug} == "yes" ]] ; then
+			if [[ $gentoo_bug = yes || $always_overflow = yes ]] ; then
 				die "install aborted due to" \
 					"poor programming practices shown above"
 			else
@@ -490,17 +672,6 @@ install_qa_check() {
 		fi
 	fi
 
-	# Compiled python objects do not belong in /usr/share (FHS violation)
-	# and can be a pain when upgrading python
-	f=$([ -d "${D}"/usr/share ] && \
-		find "${D}"usr/share -name '*.py[co]' | sed "s:${D}:/:")
-	if [[ -n ${f} ]] ; then
-		vecho -ne '\a\n'
-		eqawarn "QA Notice: Precompiled python object files do not belong in /usr/share"
-		eqawarn "${f}"
-		vecho -ne '\a\n'
-	fi
-
 	# Portage regenerates this on the installed system.
 	rm -f "${D}"/usr/share/info/dir{,.gz,.bz2}
 
@@ -508,7 +679,7 @@ install_qa_check() {
 	   [[ -x /usr/bin/file && -x /usr/bin/find ]] && \
 	   [[ -n ${MULTILIB_STRICT_DIRS} && -n ${MULTILIB_STRICT_DENY} ]]
 	then
-		local abort=no firstrun=yes
+		local abort=no dir file firstrun=yes
 		MULTILIB_STRICT_EXEMPT=$(echo ${MULTILIB_STRICT_EXEMPT} | sed -e 's:\([(|)]\):\\\1:g')
 		for dir in ${MULTILIB_STRICT_DIRS} ; do
 			[[ -d ${D}/${dir} ]] || continue
@@ -536,6 +707,7 @@ install_mask() {
 	# we don't want globbing for initial expansion, but afterwards, we do
 	local shopts=$-
 	set -o noglob
+	local no_inst
 	for no_inst in ${install_mask}; do
 		set +o noglob
 		quiet_mode || einfo "Removing ${no_inst}"
@@ -578,6 +750,7 @@ preinst_mask() {
 	cd "${T}"
 
 	# remove man pages, info pages, docs if requested
+	local f
 	for f in man info doc; do
 		if hasq no${f} $FEATURES; then
 			INSTALL_MASK="${INSTALL_MASK} /usr/share/${f}"
@@ -634,7 +807,7 @@ preinst_suid_scan() {
 	fi
 	# total suid control.
 	if hasq suidctl $FEATURES; then
-		local sfconf
+		local i sfconf x
 		sfconf=${PORTAGE_CONFIGROOT}etc/portage/suidctl.conf
 		# sandbox prevents us from writing directly
 		# to files outside of the sandbox, but this
@@ -648,8 +821,7 @@ preinst_suid_scan() {
 					vecho "- ${install_path} is an approved suid file"
 				else
 					vecho ">>> Removing sbit on non registered ${install_path}"
-					for x in 5 4 3 2 1 0; do echo -ne "\a"; sleep 0.25 ; done
-					vecho -ne "\a"
+					for x in 5 4 3 2 1 0; do sleep 0.25 ; done
 					ls_ret=$(ls -ldh "${i}")
 					chmod ugo-s "${i}"
 					grep "^#${install_path}$" "${sfconf}" > /dev/null || {
@@ -705,13 +877,13 @@ dyn_package() {
 	# for $PKGDIR and/or $PKGDIR/All.
 	export SANDBOX_ON="0"
 	[ -z "${PORTAGE_BINPKG_TMPFILE}" ] && \
-		PORTAGE_BINPKG_TMPFILE="${PKGDIR}/${CATEGORY}/${PF}.tbz2"
+		die "PORTAGE_BINPKG_TMPFILE is unset"
 	mkdir -p "${PORTAGE_BINPKG_TMPFILE%/*}" || die "mkdir failed"
 	tar $tar_options -cf - $PORTAGE_BINPKG_TAR_OPTS -C "${D}" . | \
-		bzip2 -cf > "$PORTAGE_BINPKG_TMPFILE"
+		$PORTAGE_BZIP2_COMMAND -c > "$PORTAGE_BINPKG_TMPFILE"
 	assert "failed to pack binary package: '$PORTAGE_BINPKG_TMPFILE'"
-	EPYTHON= PYTHONPATH=${PORTAGE_PYM_PATH}${PYTHONPATH:+:}${PYTHONPATH} \
-		"$PORTAGE_BIN_PATH"/xpak-helper.py recompose \
+	PYTHONPATH=${PORTAGE_PYM_PATH}${PYTHONPATH:+:}${PYTHONPATH} \
+		"${PORTAGE_PYTHON:-/usr/bin/python}" "$PORTAGE_BIN_PATH"/xpak-helper.py recompose \
 		"$PORTAGE_BINPKG_TMPFILE" "$PORTAGE_BUILDDIR/build-info"
 	if [ $? -ne 0 ]; then
 		rm -f "${PORTAGE_BINPKG_TMPFILE}"
@@ -735,8 +907,10 @@ dyn_package() {
 dyn_spec() {
 	local sources_dir=/usr/src/rpm/SOURCES
 	mkdir -p "${sources_dir}"
+	declare -a tar_args=("${EBUILD}")
+	[[ -d ${FILESDIR} ]] && tar_args=("${EBUILD}" "${FILESDIR}")
 	tar czf "${sources_dir}/${PF}.tar.gz" \
-		"${EBUILD}" "${FILESDIR}" || \
+		"${tar_args[@]}" || \
 		die "Failed to create base rpm tarball."
 
 	cat <<__END1__ > ${PF}.spec
@@ -803,9 +977,12 @@ if [ -n "${MISC_FUNCTIONS_ARGS}" ]; then
 	for x in ${MISC_FUNCTIONS_ARGS}; do
 		${x}
 	done
+	unset x
+	[[ -n $PORTAGE_EBUILD_EXIT_FILE ]] && > "$PORTAGE_EBUILD_EXIT_FILE"
+	if [[ -n $PORTAGE_IPC_DAEMON ]] ; then
+		[[ ! -s $SANDBOX_LOG ]]
+		"$PORTAGE_BIN_PATH"/ebuild-ipc exit $?
+	fi
 fi
 
-[ -n "${EBUILD_EXIT_STATUS_FILE}" ] && \
-	touch "${EBUILD_EXIT_STATUS_FILE}" &>/dev/null
-
 :
