@@ -1,31 +1,37 @@
-# Copyright 1999-2009 Gentoo Foundation
+# Copyright 1999-2010 Gentoo Foundation
 # Distributed under the terms of the GNU General Public License v2
-# $Id: Scheduler.py 15708 2010-03-02 21:13:39Z zmedico $
 
 from __future__ import print_function
 
-import codecs
+import gc
+import gzip
 import logging
+import shutil
 import sys
+import tempfile
 import textwrap
 import time
+import warnings
 import weakref
+import zlib
 
 import portage
 from portage import StringIO
 from portage import os
 from portage import _encodings
-from portage import _unicode_decode
-from portage import _unicode_encode
+from portage import _unicode_decode, _unicode_encode
 from portage.cache.mappings import slot_dict_class
+from portage.const import LIBC_PACKAGE_ATOM
 from portage.elog.messages import eerror
-from portage.output import colorize, create_color_func, darkgreen, red
+from portage.localization import _
+from portage.output import colorize, create_color_func, red
 bad = create_color_func("BAD")
 from portage._sets import SETPREFIX
 from portage._sets.base import InternalPackageSet
 from portage.util import writemsg, writemsg_level
 from portage.package.ebuild.digestcheck import digestcheck
 from portage.package.ebuild.digestgen import digestgen
+from portage.package.ebuild.prepare_build_dirs import prepare_build_dirs
 
 from _emerge.BinpkgPrefetcher import BinpkgPrefetcher
 from _emerge.Blocker import Blocker
@@ -38,6 +44,7 @@ from _emerge.depgraph import depgraph, resume_depgraph
 from _emerge.EbuildFetcher import EbuildFetcher
 from _emerge.EbuildPhase import EbuildPhase
 from _emerge.emergelog import emergelog, _emerge_log_dir
+from _emerge.FakeVartree import FakeVartree
 from _emerge._find_deep_system_runtime_deps import _find_deep_system_runtime_deps
 from _emerge._flush_elog_mod_echo import _flush_elog_mod_echo
 from _emerge.JobStatusDisplay import JobStatusDisplay
@@ -55,6 +62,9 @@ if sys.hexversion >= 0x3000000:
 
 class Scheduler(PollScheduler):
 
+	# max time between display status updates (milliseconds)
+	_max_display_latency = 3000
+
 	_opts_ignore_blockers = \
 		frozenset(["--buildpkgonly",
 		"--fetchonly", "--fetch-all-uri",
@@ -74,7 +84,8 @@ class Scheduler(PollScheduler):
 
 	class _iface_class(SlotObject):
 		__slots__ = ("dblinkEbuildPhase", "dblinkDisplayMerge",
-			"dblinkElog", "dblinkEmergeLog", "fetch", "register", "schedule",
+			"dblinkElog", "dblinkEmergeLog", "fetch",
+			"output", "register", "schedule",
 			"scheduleSetup", "scheduleUnpack", "scheduleYield",
 			"unregister")
 
@@ -130,15 +141,21 @@ class Scheduler(PollScheduler):
 			portage.exception.PortageException.__init__(self, value)
 
 	def __init__(self, settings, trees, mtimedb, myopts,
-		spinner, mergelist, favorites, digraph):
+		spinner, mergelist=None, favorites=None, graph_config=None):
 		PollScheduler.__init__(self)
+
+		if mergelist is not None:
+			warnings.warn("The mergelist parameter of the " + \
+				"_emerge.Scheduler constructor is now unused. Use " + \
+				"the graph_config parameter instead.",
+				DeprecationWarning, stacklevel=2)
+
 		self.settings = settings
 		self.target_root = settings["ROOT"]
 		self.trees = trees
 		self.myopts = myopts
 		self._spinner = spinner
 		self._mtimedb = mtimedb
-		self._mergelist = mergelist
 		self._favorites = favorites
 		self._args_set = InternalPackageSet(favorites)
 		self._build_opts = self._build_opts_class()
@@ -193,10 +210,8 @@ class Scheduler(PollScheduler):
 			self.edebug = 1
 		self.pkgsettings = {}
 		self._config_pool = {}
-		self._blocker_db = {}
-		for root in trees:
+		for root in self.trees:
 			self._config_pool[root] = []
-			self._blocker_db[root] = BlockerDB(trees[root]["root_config"])
 
 		fetch_iface = self._fetch_iface_class(log_file=self._fetch_log,
 			schedule=self._schedule_fetch)
@@ -205,7 +220,8 @@ class Scheduler(PollScheduler):
 			dblinkDisplayMerge=self._dblink_display_merge,
 			dblinkElog=self._dblink_elog,
 			dblinkEmergeLog=self._dblink_emerge_log,
-			fetch=fetch_iface, register=self._register,
+			fetch=fetch_iface, output=self._task_output,
+			register=self._register,
 			schedule=self._schedule_wait,
 			scheduleSetup=self._schedule_setup,
 			scheduleUnpack=self._schedule_unpack,
@@ -214,6 +230,7 @@ class Scheduler(PollScheduler):
 
 		self._prefetchers = weakref.WeakValueDictionary()
 		self._pkg_queue = []
+		self._running_tasks = set()
 		self._completed_tasks = set()
 
 		self._failed_pkgs = []
@@ -221,7 +238,8 @@ class Scheduler(PollScheduler):
 		self._failed_pkgs_die_msgs = []
 		self._post_mod_echo_msgs = []
 		self._parallel_fetch = False
-		merge_count = len([x for x in mergelist \
+		self._init_graph(graph_config)
+		merge_count = len([x for x in self._mergelist \
 			if isinstance(x, Package) and x.operation == "merge"])
 		self._pkg_count = self._pkg_count_class(
 			curval=0, maxval=merge_count)
@@ -235,8 +253,6 @@ class Scheduler(PollScheduler):
 		self._job_delay_exp = 1.5
 		self._previous_job_start_time = None
 
-		self._set_digraph(digraph)
-
 		# This is used to memoize the _choose_pkg() result when
 		# no packages can be chosen until one of the existing
 		# jobs completes.
@@ -256,7 +272,7 @@ class Scheduler(PollScheduler):
 					"thus parallel-fetching is being disabled"+"\n",
 					noiselevel=-1)
 				portage.writemsg(red("!!!")+"\n", noiselevel=-1)
-			elif len(mergelist) > 1:
+			elif merge_count > 1:
 				self._parallel_fetch = True
 
 		if self._parallel_fetch:
@@ -274,9 +290,80 @@ class Scheduler(PollScheduler):
 			self._running_portage = self._pkg(cpv, "installed",
 				self._running_root, installed=True)
 
+	def _init_graph(self, graph_config):
+		"""
+		Initialization structures used for dependency calculations
+		involving currently installed packages.
+		"""
+		# TODO: Replace the BlockerDB with a depgraph of installed packages
+		# that's updated incrementally with each upgrade/uninstall operation
+		# This will be useful for making quick and safe decisions with respect
+		# to aggressive parallelization discussed in bug #279623.
+		self._set_graph_config(graph_config)
+		self._blocker_db = {}
+		for root in self.trees:
+			if graph_config is None:
+				fake_vartree = FakeVartree(self.trees[root]["root_config"])
+			else:
+				fake_vartree = graph_config.trees[root]['vartree']
+			self._blocker_db[root] = BlockerDB(fake_vartree)
+
+	def _destroy_graph(self):
+		"""
+		Use this to free memory at the beginning of _calc_resume_list().
+		After _calc_resume_list(), the _init_graph() method
+		must to be called in order to re-generate the structures that
+		this method destroys. 
+		"""
+		self._blocker_db = None
+		self._set_graph_config(None)
+		gc.collect()
+
 	def _poll(self, timeout=None):
+
 		self._schedule()
-		PollScheduler._poll(self, timeout=timeout)
+
+		if timeout is None:
+			while True:
+				if not self._poll_event_handlers:
+					self._schedule()
+					if not self._poll_event_handlers:
+						raise StopIteration(
+							"timeout is None and there are no poll() event handlers")
+				previous_count = len(self._poll_event_queue)
+				PollScheduler._poll(self, timeout=self._max_display_latency)
+				self._status_display.display()
+				if previous_count != len(self._poll_event_queue):
+					break
+
+		elif timeout <= self._max_display_latency:
+			PollScheduler._poll(self, timeout=timeout)
+			if timeout == 0:
+				# The display is updated by _schedule() above, so it would be
+				# redundant to update it here when timeout is 0.
+				pass
+			else:
+				self._status_display.display()
+
+		else:
+			remaining_timeout = timeout
+			start_time = time.time()
+			while True:
+				previous_count = len(self._poll_event_queue)
+				PollScheduler._poll(self,
+					timeout=min(self._max_display_latency, remaining_timeout))
+				self._status_display.display()
+				if previous_count != len(self._poll_event_queue):
+					break
+				elapsed_time = time.time() - start_time
+				if elapsed_time < 0:
+					# The system clock has changed such that start_time
+					# is now in the future, so just assume that the
+					# timeout has already elapsed.
+					break
+				remaining_timeout = timeout - 1000 * elapsed_time
+				if remaining_timeout <= 0:
+					break
 
 	def _set_max_jobs(self, max_jobs):
 		self._max_jobs = max_jobs
@@ -340,17 +427,36 @@ class Scheduler(PollScheduler):
 				interactive_tasks.append(task)
 		return interactive_tasks
 
-	def _set_digraph(self, digraph):
+	def _set_graph_config(self, graph_config):
+
+		if graph_config is None:
+			self._graph_config = None
+			self._digraph = None
+			self._mergelist = []
+			self._deep_system_deps.clear()
+			return
+
+		self._graph_config = graph_config
+		self._digraph = graph_config.graph
+		self._mergelist = graph_config.mergelist
+
 		if "--nodeps" in self.myopts or \
 			(self._max_jobs is not True and self._max_jobs < 2):
 			# save some memory
 			self._digraph = None
+			graph_config.graph = None
+			graph_config.pkg_cache.clear()
+			self._deep_system_deps.clear()
 			return
 
-		self._digraph = digraph
 		self._find_system_deps()
 		self._prune_digraph()
 		self._prevent_builddir_collisions()
+		self._implicit_libc_deps()
+		if '--debug' in self.myopts:
+			writemsg("\nscheduler digraph:\n\n", noiselevel=-1)
+			self._digraph.debug_print()
+			writemsg("\n", noiselevel=-1)
 
 	def _find_system_deps(self):
 		"""
@@ -412,6 +518,69 @@ class Scheduler(PollScheduler):
 					priority=DepPriority(buildtime=True))
 			cpv_map[pkg.cpv].append(pkg)
 
+	def _implicit_libc_deps(self):
+		"""
+		Create implicit dependencies on libc, in order to ensure that libc
+		is installed as early as possible (see bug #303567). If the merge
+		list contains both a new-style virtual and an old-style PROVIDE
+		virtual, the new-style virtual is used.
+		"""
+		implicit_libc_roots = set([self._running_root.root])
+		libc_set = InternalPackageSet([LIBC_PACKAGE_ATOM])
+		norm_libc_pkgs = {}
+		virt_libc_pkgs = {}
+		for pkg in self._mergelist:
+			if not isinstance(pkg, Package):
+				# a satisfied blocker
+				continue
+			if pkg.installed:
+				continue
+			if pkg.root in implicit_libc_roots and \
+				pkg.operation == 'merge':
+				if libc_set.findAtomForPackage(pkg):
+					if pkg.category == 'virtual':
+						d = virt_libc_pkgs
+					else:
+						d = norm_libc_pkgs
+					if pkg.root in d:
+						raise AssertionError(
+							"found 2 libc matches: %s and %s" % \
+							(d[pkg.root], pkg))
+					d[pkg.root] = pkg
+
+		# Prefer new-style virtuals over old-style PROVIDE virtuals.
+		libc_pkg_map = norm_libc_pkgs.copy()
+		libc_pkg_map.update(virt_libc_pkgs)
+
+		# Only add a dep when the version changes.
+		for libc_pkg in list(libc_pkg_map.values()):
+			if libc_pkg.root_config.trees['vartree'].dbapi.cpv_exists(
+				libc_pkg.cpv):
+				del libc_pkg_map[pkg.root]
+
+		if not libc_pkg_map:
+			return
+
+		libc_pkgs = set(libc_pkg_map.values())
+		earlier_libc_pkgs = set()
+
+		for pkg in self._mergelist:
+			if not isinstance(pkg, Package):
+				# a satisfied blocker
+				continue
+			if pkg.installed:
+				continue
+			if pkg.root in implicit_libc_roots and \
+				pkg.operation == 'merge':
+				if pkg in libc_pkgs:
+					earlier_libc_pkgs.add(pkg)
+				else:
+					my_libc = libc_pkg_map.get(pkg.root)
+					if my_libc is not None and \
+						my_libc in earlier_libc_pkgs:
+						self._digraph.add(my_libc, pkg,
+							priority=DepPriority(buildtime=True))
+
 	class _pkg_failure(portage.exception.PortageException):
 		"""
 		An instance of this class is raised by unmerge() when
@@ -461,7 +630,6 @@ class Scheduler(PollScheduler):
 		# Call gc.collect() here to avoid heap overflow that
 		# triggers 'Cannot allocate memory' errors (reported
 		# with python-2.5).
-		import gc
 		gc.collect()
 
 		blocker_db = self._blocker_db[new_pkg.root]
@@ -489,22 +657,9 @@ class Scheduler(PollScheduler):
 		installed = type_name == "installed"
 		return self._pkg(cpv, type_name, root_config, installed=installed)
 
-	def _append_to_log_path(self, log_path, msg):
-
-		f = codecs.open(_unicode_encode(log_path,
-			encoding=_encodings['fs'], errors='strict'),
-			mode='a', encoding=_encodings['content'],
-			errors='backslashreplace')
-		try:
-			f.write(_unicode_decode(msg,
-				encoding=_encodings['content'], errors='replace'))
-		finally:
-			f.close()
-
 	def _dblink_elog(self, pkg_dblink, phase, func, msgs):
 
 		log_path = pkg_dblink.settings.get("PORTAGE_LOG_FILE")
-		background = self._background
 		out = StringIO()
 
 		for msg in msgs:
@@ -512,11 +667,7 @@ class Scheduler(PollScheduler):
 
 		out_str = out.getvalue()
 
-		if not background:
-			portage.util.writemsg_stdout(out_str, noiselevel=-1)
-
-		if log_path is not None:
-			self._append_to_log_path(log_path, out_str)
+		self._task_output(out_str, log_path=log_path)
 
 	def _dblink_emerge_log(self, msg):
 		self._logger.log(msg)
@@ -530,10 +681,7 @@ class Scheduler(PollScheduler):
 				portage.util.writemsg_level(msg,
 					level=level, noiselevel=noiselevel)
 		else:
-			if not background:
-				portage.util.writemsg_level(msg,
-					level=level, noiselevel=noiselevel)
-			self._append_to_log_path(log_path, msg)
+			self._task_output(msg, log_path=log_path)
 
 	def _dblink_ebuild_phase(self,
 		pkg_dblink, pkg_dbapi, ebuild_path, phase):
@@ -551,12 +699,12 @@ class Scheduler(PollScheduler):
 
 		if phase in ('die_hooks', 'success_hooks'):
 			ebuild_phase = MiscFunctionsProcess(background=background,
-				commands=[phase], phase=phase, pkg=pkg,
+				commands=[phase], phase=phase,
 				scheduler=scheduler, settings=settings)
 		else:
 			ebuild_phase = EbuildPhase(background=background,
-				pkg=pkg, phase=phase, scheduler=scheduler,
-				settings=settings, tree=pkg_dblink.treetype)
+				phase=phase, scheduler=scheduler,
+				settings=settings)
 		ebuild_phase.start()
 		ebuild_phase.wait()
 
@@ -575,6 +723,10 @@ class Scheduler(PollScheduler):
 		digest = '--digest' in self.myopts
 		if not digest:
 			for pkgsettings in self.pkgsettings.values():
+				if pkgsettings.mycpv is not None:
+					# ensure that we are using global features
+					# settings rather than those from package.env
+					pkgsettings.reset()
 				if 'digest' in pkgsettings.features:
 					digest = True
 					break
@@ -588,6 +740,10 @@ class Scheduler(PollScheduler):
 				x.operation != 'merge':
 				continue
 			pkgsettings = self.pkgsettings[x.root]
+			if pkgsettings.mycpv is not None:
+				# ensure that we are using global features
+				# settings rather than those from package.env
+				pkgsettings.reset()
 			if '--digest' not in self.myopts and \
 				'digest' not in pkgsettings.features:
 				continue
@@ -604,6 +760,37 @@ class Scheduler(PollScheduler):
 
 		return os.EX_OK
 
+	def _env_sanity_check(self):
+		"""
+		Verify a sane environment before trying to build anything from source.
+		"""
+		have_src_pkg = False
+		for x in self._mergelist:
+			if isinstance(x, Package) and not x.built:
+				have_src_pkg = True
+				break
+
+		if not have_src_pkg:
+			return os.EX_OK
+
+		for settings in self.pkgsettings.values():
+			for var in ("ARCH", ):
+				value = settings.get(var)
+				if value and value.strip():
+					continue
+				msg = _("%(var)s is not set... "
+					"Are you missing the '%(configroot)setc/make.profile' symlink? "
+					"Is the symlink correct? "
+					"Is your portage tree complete?") % \
+					{"var": var, "configroot": settings["PORTAGE_CONFIGROOT"]}
+
+				out = portage.output.EOutput()
+				for line in textwrap.wrap(msg, 70):
+					out.eerror(line)
+				return 1
+
+		return os.EX_OK
+
 	def _check_manifests(self):
 		# Verify all the manifests now so that the user is notified of failure
 		# as soon as possible.
@@ -728,9 +915,12 @@ class Scheduler(PollScheduler):
 		if pkg.root == self._running_root.root and \
 			portage.match_from_list(
 			portage.const.PORTAGE_PACKAGE_ATOM, [pkg]):
-			if self._running_portage:
-				return pkg.cpv != self._running_portage.cpv
-			return True
+			if self._running_portage is None:
+				return True
+			elif pkg.cpv != self._running_portage.cpv or \
+				'9999' in pkg.cpv or \
+				'git' in pkg.inherited:
+				return True
 		return False
 
 	def _restart_if_necessary(self, pkg):
@@ -778,14 +968,100 @@ class Scheduler(PollScheduler):
 			if myopt not in bad_resume_opts:
 				if myarg is True:
 					mynewargv.append(myopt)
+				elif isinstance(myarg, list):
+					# arguments like --exclude that use 'append' action
+					for x in myarg:
+						mynewargv.append("%s=%s" % (myopt, x))
 				else:
-					mynewargv.append(myopt +"="+ str(myarg))
+					mynewargv.append("%s=%s" % (myopt, myarg))
 		# priority only needs to be adjusted on the first run
 		os.environ["PORTAGE_NICENESS"] = "0"
 		os.execv(mynewargv[0], mynewargv)
 
-	def merge(self):
+	def _run_pkg_pretend(self):
+		shown_verifying_msg = False
 
+		failures = 0
+
+		for x in self._mergelist:
+			if not isinstance(x, Package):
+				continue
+
+			if x.operation == "uninstall":
+				continue
+
+			if x.metadata["EAPI"] in ("0", "1", "2", "3"):
+				continue
+
+			if "pretend" not in x.metadata.defined_phases:
+				continue
+
+			if not shown_verifying_msg and self._background:
+				shown_verifying_msg = True
+				self._status_msg("Running pre-merge checks")
+
+			if not self._background:
+				out_str =">>> Running pre-merge checks for " + colorize("INFORM", x.cpv) + "\n"
+				portage.util.writemsg_stdout(out_str, noiselevel=-1)
+
+			root_config = x.root_config
+			settings = self.pkgsettings[root_config.root]
+			tmpdir = tempfile.mkdtemp()
+			tmpdir_orig = settings["PORTAGE_TMPDIR"]
+			settings["PORTAGE_TMPDIR"] = tmpdir
+
+			if x.built:
+				tree = "bintree"
+				bintree = root_config.trees["bintree"].dbapi.bintree
+				if bintree.isremote(x.cpv):
+					fetcher = BinpkgPrefetcher(background=self._background,
+						logfile=self.settings.get("PORTAGE_LOG_FILE"), pkg=x, scheduler=self._sched_iface)
+					fetcher.start()
+					fetcher.wait()
+
+				tbz2_file = bintree.getname(x.cpv)
+				infloc = os.path.join(tmpdir, x.category, x.pf, "build-info")
+				os.makedirs(infloc)
+				portage.xpak.tbz2(tbz2_file).unpackinfo(infloc)
+				ebuild_path = os.path.join(infloc, x.pf + ".ebuild")
+
+			else:
+				tree = "porttree"
+				portdb = root_config.trees["porttree"].dbapi
+				ebuild_path = portdb.findname(x.cpv)
+				if ebuild_path is None:
+					raise AssertionError("ebuild not found for '%s'" % x.cpv)
+
+			portage.package.ebuild.doebuild.doebuild_environment(ebuild_path,
+				"pretend", settings=settings,
+				db=self.trees[settings["ROOT"]][tree].dbapi)
+			prepare_build_dirs(root_config.root, settings, cleanup=0)
+
+			vardb = root_config.trees['vartree'].dbapi
+			settings["REPLACING_VERSIONS"] = " ".join(
+				set(portage.versions.cpv_getversion(match) \
+					for match in vardb.match(x.slot_atom) + \
+					vardb.match('='+x.cpv)))
+			pretend_phase = EbuildPhase(background=self._background,
+				phase="pretend", scheduler=self._sched_iface,
+				settings=settings)
+
+			pretend_phase.start()
+			ret = pretend_phase.wait()
+
+			portage.elog.elog_process(x.cpv, settings)
+
+			if ret == os.EX_OK:
+				shutil.rmtree(tmpdir)
+			settings["PORTAGE_TMPDIR"] = tmpdir_orig
+
+			if ret != os.EX_OK:
+				failures += 1
+		if failures:
+			return 1
+		return os.EX_OK
+
+	def merge(self):
 		if "--resume" in self.myopts:
 			# We're resuming.
 			portage.writemsg_stdout(
@@ -835,12 +1111,20 @@ class Scheduler(PollScheduler):
 		if rval != os.EX_OK:
 			return rval
 
+		rval = self._env_sanity_check()
+		if rval != os.EX_OK:
+			return rval
+
 		# TODO: Immediately recalculate deps here if --keep-going
 		#       is enabled and corrupt manifests are detected.
 		rval = self._check_manifests()
 		if rval != os.EX_OK and not keep_going:
 			return rval
 
+		rval = self._run_pkg_pretend()
+		if rval != os.EX_OK:
+			return rval
+
 		while True:
 			rval = self._merge()
 			if rval == os.EX_OK or fetchonly or not keep_going:
@@ -897,16 +1181,22 @@ class Scheduler(PollScheduler):
 			log_path = self._locate_failure_log(failed_pkg)
 			if log_path is not None:
 				try:
-					log_file = codecs.open(_unicode_encode(log_path,
-					encoding=_encodings['fs'], errors='strict'),
-					mode='r', encoding=_encodings['content'], errors='replace')
+					log_file = open(_unicode_encode(log_path,
+						encoding=_encodings['fs'], errors='strict'), mode='rb')
 				except IOError:
 					pass
+				else:
+					if log_path.endswith('.gz'):
+						log_file =  gzip.GzipFile(filename='',
+							mode='rb', fileobj=log_file)
 
 			if log_file is not None:
 				try:
 					for line in log_file:
 						writemsg_level(line, noiselevel=-1)
+				except zlib.error as e:
+					writemsg_level("%s\n" % (e,), level=logging.ERROR,
+						noiselevel=-1)
 				finally:
 					log_file.close()
 				failure_log_shown = True
@@ -943,7 +1233,7 @@ class Scheduler(PollScheduler):
 				msg()
 
 		if len(self._failed_pkgs_all) > 1 or \
-			(self._failed_pkgs_all and "--keep-going" in self.myopts):
+			(self._failed_pkgs_all and keep_going):
 			if len(self._failed_pkgs_all) > 1:
 				msg = "The following %d packages have " % \
 					len(self._failed_pkgs_all) + \
@@ -957,7 +1247,9 @@ class Scheduler(PollScheduler):
 				printer.eerror(line)
 			printer.eerror("")
 			for failed_pkg in self._failed_pkgs_all:
-				msg = " %s" % (colorize('INFORM', failed_pkg.pkg.__str__()),)
+				# Use _unicode_decode() to force unicode format string so
+				# that Package.__unicode__() is called in python2.
+				msg = _unicode_decode(" %s") % (failed_pkg.pkg,)
 				log_path = self._locate_failure_log(failed_pkg)
 				if log_path is not None:
 					msg += ", Log file:"
@@ -1063,6 +1355,7 @@ class Scheduler(PollScheduler):
 
 	def _do_merge_exit(self, merge):
 		pkg = merge.merge.pkg
+		self._running_tasks.remove(pkg)
 		if merge.returncode != os.EX_OK:
 			settings = merge.merge.settings
 			build_dir = settings.get("PORTAGE_BUILDDIR")
@@ -1115,6 +1408,7 @@ class Scheduler(PollScheduler):
 				self._task_queues.merge.add(merge)
 				self._status_display.merges = len(self._task_queues.merge)
 		else:
+			self._running_tasks.remove(build.pkg)
 			settings = build.settings
 			build_dir = settings.get("PORTAGE_BUILDDIR")
 			build_log = settings.get("PORTAGE_LOG_FILE")
@@ -1179,14 +1473,14 @@ class Scheduler(PollScheduler):
 			return None
 
 		if self._digraph is None:
-			if (self._jobs or self._task_queues.merge) and \
+			if self._is_work_scheduled() and \
 				not ("--nodeps" in self.myopts and \
 				(self._max_jobs is True or self._max_jobs > 1)):
 				self._choose_pkg_return_early = True
 				return None
 			return self._pkg_queue.pop(0)
 
-		if not (self._jobs or self._task_queues.merge):
+		if not self._is_work_scheduled():
 			return self._pkg_queue.pop(0)
 
 		self._prune_digraph()
@@ -1255,7 +1549,11 @@ class Scheduler(PollScheduler):
 				node in later):
 				dependent = True
 				break
-			node_stack.extend(graph.child_nodes(node))
+
+			# Don't traverse children of uninstall nodes since
+			# those aren't dependencies in the usual sense.
+			if node.operation != "uninstall":
+				node_stack.extend(graph.child_nodes(node))
 
 		return dependent
 
@@ -1286,15 +1584,13 @@ class Scheduler(PollScheduler):
 			self._opts_no_background.intersection(self.myopts):
 			self._set_max_jobs(1)
 
-		merge_queue = self._task_queues.merge
-
 		while self._schedule():
 			if self._poll_event_handlers:
 				self._poll_loop()
 
 		while True:
 			self._schedule()
-			if not (self._jobs or merge_queue):
+			if not self._is_work_scheduled():
 				break
 			if self._poll_event_handlers:
 				self._poll_loop()
@@ -1303,37 +1599,42 @@ class Scheduler(PollScheduler):
 		return bool(self._pkg_queue and \
 			not (self._failed_pkgs and not self._build_opts.fetchonly))
 
+	def _is_work_scheduled(self):
+		return bool(self._running_tasks)
+
 	def _schedule_tasks(self):
 
-		# When the number of jobs drops to zero, process all waiting merges.
-		if not self._jobs and self._merge_wait_queue:
-			for task in self._merge_wait_queue:
-				task.addExitListener(self._merge_wait_exit_handler)
-				self._task_queues.merge.add(task)
-			self._status_display.merges = len(self._task_queues.merge)
-			self._merge_wait_scheduled.extend(self._merge_wait_queue)
-			del self._merge_wait_queue[:]
+		while True:
 
-		self._schedule_tasks_imp()
-		self._status_display.display()
+			# When the number of jobs drops to zero, process all waiting merges.
+			if not self._jobs and self._merge_wait_queue:
+				for task in self._merge_wait_queue:
+					task.addExitListener(self._merge_wait_exit_handler)
+					self._task_queues.merge.add(task)
+				self._status_display.merges = len(self._task_queues.merge)
+				self._merge_wait_scheduled.extend(self._merge_wait_queue)
+				del self._merge_wait_queue[:]
 
-		state_change = 0
-		for q in self._task_queues.values():
-			if q.schedule():
-				state_change += 1
-
-		# Cancel prefetchers if they're the only reason
-		# the main poll loop is still running.
-		if self._failed_pkgs and not self._build_opts.fetchonly and \
-			not (self._jobs or self._task_queues.merge) and \
-			self._task_queues.fetch:
-			self._task_queues.fetch.clear()
-			state_change += 1
-
-		if state_change:
 			self._schedule_tasks_imp()
 			self._status_display.display()
 
+			state_change = 0
+			for q in self._task_queues.values():
+				if q.schedule():
+					state_change += 1
+
+			# Cancel prefetchers if they're the only reason
+			# the main poll loop is still running.
+			if self._failed_pkgs and not self._build_opts.fetchonly and \
+				not self._is_work_scheduled() and \
+				self._task_queues.fetch:
+				self._task_queues.fetch.clear()
+				state_change += 1
+
+			if not (state_change or \
+				(not self._jobs and self._merge_wait_queue)):
+				break
+
 		return self._keep_scheduling()
 
 	def _job_delay(self):
@@ -1384,6 +1685,7 @@ class Scheduler(PollScheduler):
 				self._pkg_count.curval += 1
 
 			task = self._task(pkg)
+			self._running_tasks.add(pkg)
 
 			if pkg.installed:
 				merge = PackageMerge(merge=task)
@@ -1495,6 +1797,10 @@ class Scheduler(PollScheduler):
 		"""
 		print(colorize("GOOD", "*** Resuming merge..."))
 
+		# free some memory before creating
+		# the resume depgraph
+		self._destroy_graph()
+
 		myparams = create_depgraph_params(self.myopts, None)
 		success = False
 		e = None
@@ -1555,12 +1861,7 @@ class Scheduler(PollScheduler):
 			self._post_mod_echo_msgs.append(mydepgraph.display_problems)
 			return False
 		mydepgraph.display_problems()
-
-		mylist = mydepgraph.altlist()
-		mydepgraph.break_refs(mylist)
-		mydepgraph.break_refs(dropped_tasks)
-		self._mergelist = mylist
-		self._set_digraph(mydepgraph.schedulerGraph())
+		self._init_graph(mydepgraph.schedulerGraph())
 
 		msg_width = 75
 		for task in dropped_tasks:
