@@ -1,6 +1,5 @@
-# Copyright 1999-2009 Gentoo Foundation
+# Copyright 1999-2010 Gentoo Foundation
 # Distributed under the terms of the GNU General Public License v2
-# $Id: main.py 15687 2010-03-02 21:07:53Z zmedico $
 
 from __future__ import print_function
 
@@ -25,16 +24,15 @@ from portage.output import create_color_func
 good = create_color_func("GOOD")
 bad = create_color_func("BAD")
 
+from portage.const import _ENABLE_DYN_LINK_MAP
 import portage.elog
-import portage.dep
-portage.dep._dep_check_strict = True
 import portage.util
 import portage.locks
 import portage.exception
 from portage.data import secpass
 from portage.dbapi.dep_expand import dep_expand
 from portage.util import normalize_path as normpath
-from portage.util import writemsg, writemsg_level, writemsg_stdout
+from portage.util import shlex_split, writemsg_level, writemsg_stdout
 from portage._sets import SETPREFIX
 from portage._global_updates import _global_updates
 
@@ -54,7 +52,7 @@ if sys.hexversion >= 0x3000000:
 options=[
 "--ask",          "--alphabetical",
 "--ask-enter-invalid",
-"--buildpkg",     "--buildpkgonly",
+"--buildpkgonly",
 "--changed-use",
 "--changelog",    "--columns",
 "--debug",
@@ -82,7 +80,7 @@ options=[
 shortmapping={
 "1":"--oneshot",
 "a":"--ask",
-"b":"--buildpkg",  "B":"--buildpkgonly",
+"B":"--buildpkgonly",
 "c":"--depclean",
 "C":"--unmerge",
 "d":"--debug",
@@ -110,11 +108,13 @@ def chk_updated_info_files(root, infodirs, prev_mtimes, retval):
 			if z=='':
 				continue
 			inforoot=normpath(root+z)
-			if os.path.isdir(inforoot):
-				infomtime = os.stat(inforoot)[stat.ST_MTIME]
-				if inforoot not in prev_mtimes or \
-					prev_mtimes[inforoot] != infomtime:
-						regen_infodirs.append(inforoot)
+			if os.path.isdir(inforoot) and \
+				not [x for x in os.listdir(inforoot) \
+				if x.startswith('.keepinfodir')]:
+					infomtime = os.stat(inforoot)[stat.ST_MTIME]
+					if inforoot not in prev_mtimes or \
+						prev_mtimes[inforoot] != infomtime:
+							regen_infodirs.append(inforoot)
 
 		if not regen_infodirs:
 			portage.writemsg_stdout("\n")
@@ -211,6 +211,87 @@ def chk_updated_info_files(root, infodirs, prev_mtimes, retval):
 				if icount > 0:
 					out.einfo("Processed %d info files." % (icount,))
 
+def display_preserved_libs(vardbapi, myopts):
+	MAX_DISPLAY = 3
+
+	if vardbapi._linkmap is None or \
+		vardbapi._plib_registry is None:
+		# preserve-libs is entirely disabled
+		return
+
+	# Ensure the registry is consistent with existing files.
+	vardbapi._plib_registry.pruneNonExisting()
+
+	if vardbapi._plib_registry.hasEntries():
+		if "--quiet" in myopts:
+			print()
+			print(colorize("WARN", "!!!") + " existing preserved libs found")
+			return
+		else:
+			print()
+			print(colorize("WARN", "!!!") + " existing preserved libs:")
+
+		plibdata = vardbapi._plib_registry.getPreservedLibs()
+		linkmap = vardbapi._linkmap
+		consumer_map = {}
+		owners = {}
+		linkmap_broken = False
+
+		try:
+			linkmap.rebuild()
+		except portage.exception.CommandNotFound as e:
+			writemsg_level("!!! Command Not Found: %s\n" % (e,),
+				level=logging.ERROR, noiselevel=-1)
+			del e
+			linkmap_broken = True
+		else:
+			search_for_owners = set()
+			for cpv in plibdata:
+				internal_plib_keys = set(linkmap._obj_key(f) \
+					for f in plibdata[cpv])
+				for f in plibdata[cpv]:
+					if f in consumer_map:
+						continue
+					consumers = []
+					for c in linkmap.findConsumers(f):
+						# Filter out any consumers that are also preserved libs
+						# belonging to the same package as the provider.
+						if linkmap._obj_key(c) not in internal_plib_keys:
+							consumers.append(c)
+					consumers.sort()
+					consumer_map[f] = consumers
+					search_for_owners.update(consumers[:MAX_DISPLAY+1])
+
+			owners = vardbapi._owners.getFileOwnerMap(search_for_owners)
+
+		for cpv in plibdata:
+			print(colorize("WARN", ">>>") + " package: %s" % cpv)
+			samefile_map = {}
+			for f in plibdata[cpv]:
+				obj_key = linkmap._obj_key(f)
+				alt_paths = samefile_map.get(obj_key)
+				if alt_paths is None:
+					alt_paths = set()
+					samefile_map[obj_key] = alt_paths
+				alt_paths.add(f)
+
+			for alt_paths in samefile_map.values():
+				alt_paths = sorted(alt_paths)
+				for p in alt_paths:
+					print(colorize("WARN", " * ") + " - %s" % (p,))
+				f = alt_paths[0]
+				consumers = consumer_map.get(f, [])
+				for c in consumers[:MAX_DISPLAY]:
+					print(colorize("WARN", " * ") + "     used by %s (%s)" % \
+						(c, ", ".join(x.mycpv for x in owners.get(c, []))))
+				if len(consumers) == MAX_DISPLAY + 1:
+					print(colorize("WARN", " * ") + "     used by %s (%s)" % \
+						(consumers[MAX_DISPLAY], ", ".join(x.mycpv \
+						for x in owners.get(consumers[MAX_DISPLAY], []))))
+				elif len(consumers) > MAX_DISPLAY:
+					print(colorize("WARN", " * ") + "     used by %d other files" % (len(consumers) - MAX_DISPLAY))
+		print("Use " + colorize("GOOD", "emerge @preserved-rebuild") + " to rebuild packages using these libraries")
+
 def post_emerge(root_config, myopts, mtimedb, retval):
 	"""
 	Misc. things to run at the end of a merge session.
@@ -245,7 +326,7 @@ def post_emerge(root_config, myopts, mtimedb, retval):
 	settings.regenerate()
 	settings.lock()
 
-	config_protect = settings.get("CONFIG_PROTECT","").split()
+	config_protect = shlex_split(settings.get("CONFIG_PROTECT", ""))
 	infodirs = settings.get("INFOPATH","").split(":") + \
 		settings.get("INFODIR","").split(":")
 
@@ -264,7 +345,7 @@ def post_emerge(root_config, myopts, mtimedb, retval):
 		# If vdb state has not changed then there's nothing else to do.
 		sys.exit(retval)
 
-	vdb_path = os.path.join(target_root, portage.VDB_PATH)
+	vdb_path = os.path.join(root_config.settings['EROOT'], portage.VDB_PATH)
 	portage.util.ensure_dirs(vdb_path)
 	vdb_lock = None
 	if os.access(vdb_path, os.W_OK) and not "--pretend" in myopts:
@@ -280,9 +361,11 @@ def post_emerge(root_config, myopts, mtimedb, retval):
 			if vdb_lock:
 				portage.locks.unlockdir(vdb_lock)
 
-	chk_updated_cfg_files(target_root, config_protect)
+	chk_updated_cfg_files(settings['EROOT'], config_protect)
 
 	display_news_notification(root_config, myopts)
+	if retval in (None, os.EX_OK) or (not "--pretend" in myopts):
+		display_preserved_libs(vardbapi, myopts)	
 
 	sys.exit(retval)
 
@@ -307,28 +390,35 @@ def insert_optional_args(args):
 				return False
 
 	valid_integers = valid_integers()
+	y_or_n = ('y', 'n',)
 
 	new_args = []
 
 	default_arg_opts = {
-		'--complete-graph' : ('n',),
+		'--autounmask'           : y_or_n,
+		'--buildpkg'             : y_or_n,
+		'--complete-graph'       : y_or_n,
 		'--deep'       : valid_integers,
-		'--deselect'   : ('n',),
-		'--binpkg-respect-use'   : ('n', 'y',),
-		'--fail-clean'           : ('n',),
-		'--getbinpkg'            : ('n',),
-		'--getbinpkgonly'        : ('n',),
+		'--deselect'             : y_or_n,
+		'--binpkg-respect-use'   : y_or_n,
+		'--fail-clean'           : y_or_n,
+		'--getbinpkg'            : y_or_n,
+		'--getbinpkgonly'        : y_or_n,
 		'--jobs'       : valid_integers,
-		'--keep-going'           : ('n',),
-		'--rebuilt-binaries'     : ('n',),
+		'--keep-going'           : y_or_n,
+		'--package-moves'        : y_or_n,
+		'--rebuilt-binaries'     : y_or_n,
 		'--root-deps'  : ('rdeps',),
-		'--select'               : ('n',),
-		'--selective'            : ('n',),
-		"--use-ebuild-visibility": ('n',),
-		'--usepkg'               : ('n',),
-		'--usepkgonly'           : ('n',),
+		'--select'               : y_or_n,
+		'--selective'            : y_or_n,
+		"--use-ebuild-visibility": y_or_n,
+		'--usepkg'               : y_or_n,
+		'--usepkgonly'           : y_or_n,
 	}
 
+	if _ENABLE_DYN_LINK_MAP:
+		default_arg_opts['--depclean-lib-check'] = y_or_n
+
 	short_arg_opts = {
 		'D' : valid_integers,
 		'j' : valid_integers,
@@ -337,10 +427,11 @@ def insert_optional_args(args):
 	# Don't make things like "-kn" expand to "-k n"
 	# since existence of -n makes it too ambiguous.
 	short_arg_opts_n = {
-		'g' : ('n',),
-		'G' : ('n',),
-		'k' : ('n',),
-		'K' : ('n',),
+		'b' : y_or_n,
+		'g' : y_or_n,
+		'G' : y_or_n,
+		'k' : y_or_n,
+		'K' : y_or_n,
 	}
 
 	arg_stack = args[:]
@@ -429,13 +520,22 @@ def parse_opts(tmpcmdline, silent=False):
 
 	actions = frozenset([
 		"clean", "config", "depclean", "help",
-		"info", "metadata",
+		"info", "list-sets", "metadata",
 		"prune", "regen",  "search",
 		"sync",  "unmerge", "version",
 	])
 
 	longopt_aliases = {"--cols":"--columns", "--skip-first":"--skipfirst"}
+	true_y_or_n = ("True", "y", "n")
+	true_y = ("True", "y")
 	argument_options = {
+
+		"--autounmask": {
+			"help"    : "automatically unmask packages",
+			"type"    : "choice",
+			"choices" : true_y_or_n
+		},
+
 		"--accept-properties": {
 			"help":"temporarily override ACCEPT_PROPERTIES",
 			"action":"store"
@@ -449,6 +549,13 @@ def parse_opts(tmpcmdline, silent=False):
 			"action" : "store"
 		},
 
+		"--buildpkg": {
+			"shortopt" : "-b",
+			"help"     : "build binary packages",
+			"type"     : "choice",
+			"choices"  : true_y_or_n
+		},
+
 		"--config-root": {
 			"help":"specify the location for portage configuration files",
 			"action":"store"
@@ -462,7 +569,7 @@ def parse_opts(tmpcmdline, silent=False):
 		"--complete-graph": {
 			"help"    : "completely account for all known dependencies",
 			"type"    : "choice",
-			"choices" : ("True", "n")
+			"choices" : true_y_or_n
 		},
 
 		"--deep": {
@@ -478,15 +585,23 @@ def parse_opts(tmpcmdline, silent=False):
 		},
 
 		"--deselect": {
-			"help"    : "remove atoms from the world file",
+			"help"    : "remove atoms/sets from the world file",
 			"type"    : "choice",
-			"choices" : ("True", "n")
+			"choices" : true_y_or_n
+		},
+
+		"--exclude": {
+			"help"   :"A space separated list of package names or slot atoms. " + \
+				"Emerge won't  install any ebuild or binary package that " + \
+				"matches any of the given package atoms.",
+
+			"action" : "append"
 		},
 
 		"--fail-clean": {
 			"help"    : "clean temp files after build failure",
 			"type"    : "choice",
-			"choices" : ("True", "n")
+			"choices" : true_y_or_n
 		},
 
 		"--jobs": {
@@ -502,7 +617,7 @@ def parse_opts(tmpcmdline, silent=False):
 		"--keep-going": {
 			"help"    : "continue as much as possible after an error",
 			"type"    : "choice",
-			"choices" : ("True", "n")
+			"choices" : true_y_or_n
 		},
 
 		"--load-average": {
@@ -529,28 +644,40 @@ def parse_opts(tmpcmdline, silent=False):
 			"help"    : "discard binary packages if their use flags \
 				don't match the current configuration",
 			"type"    : "choice",
-			"choices" : ("True", "y", "n")
+			"choices" : true_y_or_n
 		},
 
 		"--getbinpkg": {
 			"shortopt" : "-g",
 			"help"     : "fetch binary packages",
 			"type"     : "choice",
-			"choices"  : ("True", "n")
+			"choices"  : true_y_or_n
 		},
 
 		"--getbinpkgonly": {
 			"shortopt" : "-G",
 			"help"     : "fetch binary packages only",
 			"type"     : "choice",
-			"choices"  : ("True", "n")
+			"choices"  : true_y_or_n
+		},
+
+		"--package-moves": {
+			"help"     : "perform package moves when necessary",
+			"type"     : "choice",
+			"choices"  : true_y_or_n
 		},
 
 		"--rebuilt-binaries": {
 			"help"     : "replace installed packages with binary " + \
 			             "packages that have been rebuilt",
 			"type"     : "choice",
-			"choices"  : ("True", "n")
+			"choices"  : true_y_or_n
+		},
+		
+		"--rebuilt-binaries-timestamp": {
+			"help"   : "use only binaries that are newer than this " + \
+			           "timestamp for --rebuilt-binaries",
+			"action" : "store"
 		},
 
 		"--root": {
@@ -568,38 +695,45 @@ def parse_opts(tmpcmdline, silent=False):
 			"help"    : "add specified packages to the world set " + \
 			            "(inverse of --oneshot)",
 			"type"    : "choice",
-			"choices" : ("True", "n")
+			"choices" : true_y_or_n
 		},
 
 		"--selective": {
 			"help"    : "similar to the --noreplace but does not take " + \
 			            "precedence over options such as --newuse",
 			"type"    : "choice",
-			"choices" : ("True", "n")
+			"choices" : true_y_or_n
 		},
 
 		"--use-ebuild-visibility": {
 			"help"     : "use unbuilt ebuild metadata for visibility checks on built packages",
 			"type"     : "choice",
-			"choices"  : ("True", "n")
+			"choices"  : true_y_or_n
 		},
 
 		"--usepkg": {
 			"shortopt" : "-k",
 			"help"     : "use binary packages",
 			"type"     : "choice",
-			"choices"  : ("True", "n")
+			"choices"  : true_y_or_n
 		},
 
 		"--usepkgonly": {
 			"shortopt" : "-K",
 			"help"     : "use only binary packages",
 			"type"     : "choice",
-			"choices"  : ("True", "n")
+			"choices"  : true_y_or_n
 		},
 
 	}
 
+	if _ENABLE_DYN_LINK_MAP:
+		argument_options["--depclean-lib-check"] = {
+			"help"    : "check for consumers of libraries before removing them",
+			"type"    : "choice",
+			"choices" : true_y_or_n
+		}
+
 	from optparse import OptionParser
 	parser = OptionParser()
 	if parser.has_option("--help"):
@@ -630,54 +764,94 @@ def parse_opts(tmpcmdline, silent=False):
 
 	myoptions, myargs = parser.parse_args(args=tmpcmdline)
 
+	if myoptions.autounmask in true_y:
+		myoptions.autounmask = True
+
+	if myoptions.buildpkg in true_y:
+		myoptions.buildpkg = True
+	else:
+		myoptions.buildpkg = None
+
 	if myoptions.changed_use is not False:
 		myoptions.reinstall = "changed-use"
 		myoptions.changed_use = False
 
-	if myoptions.deselect == "True":
+	if myoptions.deselect in true_y:
 		myoptions.deselect = True
 
-	if myoptions.binpkg_respect_use in ("y", "True",):
+	if myoptions.binpkg_respect_use in true_y:
 		myoptions.binpkg_respect_use = True
 	else:
 		myoptions.binpkg_respect_use = None
 
-	if myoptions.complete_graph in ("y", "True",):
+	if myoptions.complete_graph in true_y:
 		myoptions.complete_graph = True
 	else:
 		myoptions.complete_graph = None
 
-	if myoptions.fail_clean == "True":
+	if _ENABLE_DYN_LINK_MAP:
+		if myoptions.depclean_lib_check in true_y:
+			myoptions.depclean_lib_check = True
+
+	if myoptions.exclude:
+		exclude = []
+		bad_atoms = []
+		for x in ' '.join(myoptions.exclude).split():
+			bad_atom = False
+			try:
+				atom = portage.dep.Atom(x, allow_wildcard=True)
+			except portage.exception.InvalidAtom:
+				try:
+					atom = portage.dep.Atom("*/"+x, allow_wildcard=True)
+				except portage.exception.InvalidAtom:
+					bad_atom = True
+			
+			if bad_atom:
+				bad_atoms.append(x)
+			else:
+				if atom.operator or atom.blocker or atom.use:
+					bad_atoms.append(x)
+				else:
+					exclude.append(atom)
+
+		if bad_atoms and not silent:
+			parser.error("Invalid Atom(s) in --exclude parameter: '%s' (only package names and slot atoms (with wildcards) allowed)\n" % \
+				(",".join(bad_atoms),))
+
+	if myoptions.fail_clean in true_y:
 		myoptions.fail_clean = True
 
-	if myoptions.getbinpkg in ("True",):
+	if myoptions.getbinpkg in true_y:
 		myoptions.getbinpkg = True
 	else:
 		myoptions.getbinpkg = None
 
-	if myoptions.getbinpkgonly in ("True",):
+	if myoptions.getbinpkgonly in true_y:
 		myoptions.getbinpkgonly = True
 	else:
 		myoptions.getbinpkgonly = None
 
-	if myoptions.keep_going in ("True",):
+	if myoptions.keep_going in true_y:
 		myoptions.keep_going = True
 	else:
 		myoptions.keep_going = None
 
-	if myoptions.rebuilt_binaries in ("True",):
+	if myoptions.package_moves in true_y:
+		myoptions.package_moves = True
+
+	if myoptions.rebuilt_binaries in true_y:
 		myoptions.rebuilt_binaries = True
 
-	if myoptions.root_deps == "True":
+	if myoptions.root_deps in true_y:
 		myoptions.root_deps = True
 
-	if myoptions.select == "True":
+	if myoptions.select in true_y:
 		myoptions.select = True
 		myoptions.oneshot = False
 	elif myoptions.select == "n":
 		myoptions.oneshot = True
 
-	if myoptions.selective == "True":
+	if myoptions.selective in true_y:
 		myoptions.selective = True
 
 	if myoptions.backtrack is not None:
@@ -690,8 +864,8 @@ def parse_opts(tmpcmdline, silent=False):
 		if backtrack < 0:
 			backtrack = None
 			if not silent:
-				writemsg("!!! Invalid --backtrack parameter: '%s'\n" % \
-					(myoptions.backtrack,), noiselevel=-1)
+				parser.error("Invalid --backtrack parameter: '%s'\n" % \
+					(myoptions.backtrack,))
 
 		myoptions.backtrack = backtrack
 
@@ -708,8 +882,8 @@ def parse_opts(tmpcmdline, silent=False):
 		if deep is not True and deep < 0:
 			deep = None
 			if not silent:
-				writemsg("!!! Invalid --deep parameter: '%s'\n" % \
-					(myoptions.deep,), noiselevel=-1)
+				parser.error("Invalid --deep parameter: '%s'\n" % \
+					(myoptions.deep,))
 
 		myoptions.deep = deep
 
@@ -727,8 +901,8 @@ def parse_opts(tmpcmdline, silent=False):
 			jobs < 1:
 			jobs = None
 			if not silent:
-				writemsg("!!! Invalid --jobs parameter: '%s'\n" % \
-					(myoptions.jobs,), noiselevel=-1)
+				parser.error("Invalid --jobs parameter: '%s'\n" % \
+					(myoptions.jobs,))
 
 		myoptions.jobs = jobs
 
@@ -741,22 +915,36 @@ def parse_opts(tmpcmdline, silent=False):
 		if load_average <= 0.0:
 			load_average = None
 			if not silent:
-				writemsg("!!! Invalid --load-average parameter: '%s'\n" % \
-					(myoptions.load_average,), noiselevel=-1)
+				parser.error("Invalid --load-average parameter: '%s'\n" % \
+					(myoptions.load_average,))
 
 		myoptions.load_average = load_average
+	
+	if myoptions.rebuilt_binaries_timestamp:
+		try:
+			rebuilt_binaries_timestamp = int(myoptions.rebuilt_binaries_timestamp)
+		except ValueError:
+			rebuilt_binaries_timestamp = -1
 
-	if myoptions.use_ebuild_visibility in ("True",):
+		if rebuilt_binaries_timestamp < 0:
+			rebuilt_binaries_timestamp = 0
+			if not silent:
+				parser.error("Invalid --rebuilt-binaries-timestamp parameter: '%s'\n" % \
+					(myoptions.rebuilt_binaries_timestamp,))
+
+		myoptions.rebuilt_binaries_timestamp = rebuilt_binaries_timestamp
+
+	if myoptions.use_ebuild_visibility in true_y:
 		myoptions.use_ebuild_visibility = True
 	else:
 		myoptions.use_ebuild_visibility = None
 
-	if myoptions.usepkg in ("True",):
+	if myoptions.usepkg in true_y:
 		myoptions.usepkg = True
 	else:
 		myoptions.usepkg = None
 
-	if myoptions.usepkgonly in ("True",):
+	if myoptions.usepkgonly in true_y:
 		myoptions.usepkgonly = True
 	else:
 		myoptions.usepkgonly = None
@@ -794,10 +982,25 @@ def parse_opts(tmpcmdline, silent=False):
 
 	return myaction, myopts, myfiles
 
+# Warn about features that may confuse users and
+# lead them to report invalid bugs.
+_emerge_features_warn = frozenset(['keeptemp', 'keepwork'])
+
 def validate_ebuild_environment(trees):
+	features_warn = set()
 	for myroot in trees:
 		settings = trees[myroot]["vartree"].settings
 		settings.validate()
+		features_warn.update(
+			_emerge_features_warn.intersection(settings.features))
+
+	if features_warn:
+		msg = "WARNING: The FEATURES variable contains one " + \
+			"or more values that should be disabled under " + \
+			"normal circumstances: %s" % " ".join(features_warn)
+		out = portage.output.EOutput()
+		for line in textwrap.wrap(msg, 65):
+			out.ewarn(line)
 
 def apply_priorities(settings):
 	ionice(settings)
@@ -842,7 +1045,7 @@ def setconfig_fallback(root_config):
 	from portage._sets.profiles import PackagesSystemSet
 	setconfig = root_config.setconfig
 	setconfig.psets['world'] = DummyPackageSet(atoms=['@selected', '@system'])
-	setconfig.psets['selected'] = WorldSelectedSet(root_config.root)
+	setconfig.psets['selected'] = WorldSelectedSet(root_config.settings['EROOT'])
 	setconfig.psets['system'] = \
 		PackagesSystemSet(root_config.settings.profiles)
 	root_config.sets = setconfig.getSets()
@@ -869,8 +1072,12 @@ def missing_sets_warning(root_config, missing_sets):
 		"missing set(s): %s" % missing_sets_str]
 	if root_config.sets:
 		msg.append("        sets defined: %s" % ", ".join(root_config.sets))
+	global_config_path = portage.const.GLOBAL_CONFIG_PATH
+	if root_config.settings['EPREFIX']:
+		global_config_path = os.path.join(root_config.settings['EPREFIX'],
+				portage.const.GLOBAL_CONFIG_PATH.lstrip(os.sep))
 	msg.append("        This usually means that '%s'" % \
-		(os.path.join(portage.const.GLOBAL_CONFIG_PATH, "sets.conf"),))
+		(os.path.join(global_config_path, "sets/portage.conf"),))
 	msg.append("        is missing or corrupt.")
 	msg.append("        Falling back to default world and system set configuration!!!")
 	for line in msg:
@@ -887,34 +1094,98 @@ def ensure_required_sets(trees):
 			setconfig_fallback(root_trees["root_config"])
 
 def expand_set_arguments(myfiles, myaction, root_config):
+	retval = os.EX_OK
+	setconfig = root_config.setconfig
 
-	if myaction != "search":
+	sets = setconfig.getSets()
 
-		world = False
-		system = False
+	# In order to know exactly which atoms/sets should be added to the
+	# world file, the depgraph performs set expansion later. It will get
+	# confused about where the atoms came from if it's not allowed to
+	# expand them itself.
+	do_not_expand = (None, )
+	newargs = []
+	for a in myfiles:
+		if a in ("system", "world"):
+			newargs.append(SETPREFIX+a)
+		else:
+			newargs.append(a)
+	myfiles = newargs
+	del newargs
+	newargs = []
 
-		for x in myfiles:
-			if x[:1] == SETPREFIX:
-				msg = []
-				msg.append("'%s' is not a valid package atom." % (x,))
-				msg.append("Please check ebuild(5) for full details.")
-				writemsg_level("".join("!!! %s\n" % line for line in msg),
-					level=logging.ERROR, noiselevel=-1)
-				return (myfiles, 1)
-			elif x == "system":
-				system = True
-			elif x == "world":
-				world = True
+	# separators for set arguments
+	ARG_START = "{"
+	ARG_END = "}"
 
-		if myaction is not None:
-			if system:
-				multiple_actions("system", myaction)
-				return (myfiles, 1)
-			elif world:
-				multiple_actions("world", myaction)
-				return (myfiles, 1)
+	for i in range(0, len(myfiles)):
+		if myfiles[i].startswith(SETPREFIX):
+			start = 0
+			end = 0
+			x = myfiles[i][len(SETPREFIX):]
+			newset = ""
+			while x:
+				start = x.find(ARG_START)
+				end = x.find(ARG_END)
+				if start > 0 and start < end:
+					namepart = x[:start]
+					argpart = x[start+1:end]
 
-	return (myfiles, os.EX_OK)
+					# TODO: implement proper quoting
+					args = argpart.split(",")
+					options = {}
+					for a in args:
+						if "=" in a:
+							k, v  = a.split("=", 1)
+							options[k] = v
+						else:
+							options[a] = "True"
+					setconfig.update(namepart, options)
+					newset += (x[:start-len(namepart)]+namepart)
+					x = x[end+len(ARG_END):]
+				else:
+					newset += x
+					x = ""
+			myfiles[i] = SETPREFIX+newset
+
+	sets = setconfig.getSets()
+
+	# display errors that occured while loading the SetConfig instance
+	for e in setconfig.errors:
+		print(colorize("BAD", "Error during set creation: %s" % e))
+
+	unmerge_actions = ("unmerge", "prune", "clean", "depclean")
+
+	for a in myfiles:
+		if a.startswith(SETPREFIX):		
+				s = a[len(SETPREFIX):]
+				if s not in sets:
+					display_missing_pkg_set(root_config, s)
+					return (None, 1)
+				setconfig.active.append(s)
+				try:
+					set_atoms = setconfig.getSetAtoms(s)
+				except portage.exception.PackageSetNotFound as e:
+					writemsg_level(("emerge: the given set '%s' " + \
+						"contains a non-existent set named '%s'.\n") % \
+						(s, e), level=logging.ERROR, noiselevel=-1)
+					return (None, 1)
+				if myaction in unmerge_actions and \
+						not sets[s].supportsOperation("unmerge"):
+					sys.stderr.write("emerge: the given set '%s' does " % s + \
+						"not support unmerge operations\n")
+					retval = 1
+				elif not set_atoms:
+					print("emerge: '%s' is an empty set" % s)
+				elif myaction not in do_not_expand:
+					newargs.extend(set_atoms)
+				else:
+					newargs.append(SETPREFIX+s)
+				for e in sets[s].errors:
+					print(e)
+		else:
+			newargs.append(a)
+	return (newargs, retval)
 
 def repo_name_check(trees):
 	missing_repo_names = set()
@@ -1018,6 +1289,7 @@ def check_procfs():
 def emerge_main():
 	global portage	# NFC why this is necessary now - genone
 	portage._disable_legacy_globals()
+	portage.dep._internal_warnings = True
 	# Disable color until we're sure that it should be enabled (after
 	# EMERGE_DEFAULT_OPTS has been parsed).
 	portage.output.havecolor = 0
@@ -1043,8 +1315,15 @@ def emerge_main():
 	if rval != os.EX_OK:
 		return rval
 
+	tmpcmdline = []
+	if "--ignore-default-opts" not in myopts:
+		tmpcmdline.extend(settings["EMERGE_DEFAULT_OPTS"].split())
+	tmpcmdline.extend(sys.argv[1:])
+	myaction, myopts, myfiles = parse_opts(tmpcmdline)
+
 	if myaction not in ('help', 'info', 'version') and \
-		_global_updates(trees, mtimedb["updates"]):
+		myopts.get('--package-moves') != 'n' and \
+		_global_updates(trees, mtimedb["updates"], quiet=("--quiet" in myopts)):
 		mtimedb.commit()
 		# Reload the whole config from scratch.
 		settings, trees, mtimedb = load_emerge_config(trees=trees)
@@ -1054,12 +1333,6 @@ def emerge_main():
 	if xterm_titles:
 		xtermTitle("emerge")
 
-	tmpcmdline = []
-	if "--ignore-default-opts" not in myopts:
-		tmpcmdline.extend(settings["EMERGE_DEFAULT_OPTS"].split())
-	tmpcmdline.extend(sys.argv[1:])
-	myaction, myopts, myfiles = parse_opts(tmpcmdline)
-
 	if "--digest" in myopts:
 		os.environ["FEATURES"] = os.environ.get("FEATURES","") + " digest"
 		# Reload the whole config from scratch so that the portdbapi internal
@@ -1070,13 +1343,26 @@ def emerge_main():
 	adjust_configs(myopts, trees)
 	apply_priorities(settings)
 
+	if myaction == 'version':
+		writemsg_stdout(getportageversion(
+			settings["PORTDIR"], settings["ROOT"],
+			settings.profile_path, settings["CHOST"],
+			trees[settings["ROOT"]]["vartree"].dbapi) + '\n', noiselevel=-1)
+		return 0
+	elif myaction == 'help':
+		_emerge.help.help(myopts, portage.output.havecolor)
+		return 0
+
 	spinner = stdout_spinner()
 	if "candy" in settings.features:
 		spinner.update = spinner.update_scroll
 
 	if "--quiet" not in myopts:
 		portage.deprecated_profile_check(settings=settings)
-		#repo_name_check(trees)
+		if portage.const._ENABLE_REPO_NAME_WARN:
+			# Bug #248603 - Disable warnings about missing
+			# repo_name entries for stable branch.
+			repo_name_check(trees)
 		repo_name_duplicate_check(trees)
 		config_protect_check(trees)
 	check_procfs()
@@ -1111,7 +1397,8 @@ def emerge_main():
 		# Freeze the portdbapi for performance (memoize all xmatch results).
 		mydb.freeze()
 
-		if "--usepkg" in myopts:
+		if myaction in ('search', None) and \
+			"--usepkg" in myopts:
 			# Populate the bintree with current --getbinpkg setting.
 			# This needs to happen before expand_set_arguments(), in case
 			# any sets use the bintree.
@@ -1143,6 +1430,9 @@ def emerge_main():
 			break
 
 	root_config = trees[settings["ROOT"]]["root_config"]
+	if myaction == "list-sets":
+		writemsg_stdout("".join("%s\n" % s for s in sorted(root_config.sets)))
+		return os.EX_OK
 
 	ensure_required_sets(trees)
 
@@ -1193,6 +1483,7 @@ def emerge_main():
 	if settings.get("PORTAGE_DEBUG", "") == "1":
 		spinner.update = spinner.update_quiet
 		portage.debug=1
+		portage.util.noiselimit = 0
 		if "python-trace" in settings.features:
 			import portage.debug
 			portage.debug.set_trace(True)
@@ -1203,15 +1494,6 @@ def emerge_main():
 			not sys.stdout.isatty():
 			spinner.update = spinner.update_basic
 
-	if myaction == 'version':
-		print(getportageversion(settings["PORTDIR"], settings["ROOT"],
-			settings.profile_path, settings["CHOST"],
-			trees[settings["ROOT"]]["vartree"].dbapi))
-		return 0
-	elif myaction == "help":
-		_emerge.help.help(myopts, portage.output.havecolor)
-		return 0
-
 	if "--debug" in myopts:
 		print("myaction", myaction)
 		print("myopts", myopts)
@@ -1232,8 +1514,7 @@ def emerge_main():
 				'prune', 'unmerge') or not \
 				(fetchonly or \
 				(buildpkgonly and secpass >= 1) or \
-				myaction in ("metadata", "regen") or \
-				(myaction == "sync" and os.access(settings["PORTDIR"], os.W_OK)))
+				myaction in ("metadata", "regen", "sync"))
 			if portage.secpass < 1 or \
 				need_superuser:
 				if need_superuser:
@@ -1353,12 +1634,21 @@ def emerge_main():
 
 		# Ensure atoms are valid before calling unmerge().
 		vardb = trees[settings["ROOT"]]["vartree"].dbapi
+		portdb = trees[settings["ROOT"]]["porttree"].dbapi
+		bindb = trees[settings["ROOT"]]["bintree"].dbapi
 		valid_atoms = []
 		for x in myfiles:
 			if is_valid_package_atom(x):
 				try:
-					valid_atoms.append(
-						dep_expand(x, mydb=vardb, settings=settings))
+					#look at the installed files first, if there is no match
+					#look at the ebuilds, since EAPI 4 allows running pkg_info
+					#on non-installed packages
+					valid_atom = dep_expand(x, mydb=vardb, settings=settings)
+					if valid_atom.cp.split("/")[0] == "null":
+						valid_atom = dep_expand(x, mydb=portdb, settings=settings)
+					if valid_atom.cp.split("/")[0] == "null" and "--usepkg" in myopts:
+						valid_atom = dep_expand(x, mydb=bindb, settings=settings)
+					valid_atoms.append(valid_atom)
 				except portage.exception.AmbiguousPackageName as e:
 					msg = "The short ebuild name \"" + x + \
 						"\" is ambiguous.  Please specify " + \
@@ -1367,7 +1657,7 @@ def emerge_main():
 					for line in textwrap.wrap(msg, 70):
 						writemsg_level("!!! %s\n" % (line,),
 							level=logging.ERROR, noiselevel=-1)
-					for i in e[0]:
+					for i in e.args[0]:
 						writemsg_level("    %s\n" % colorize("INFORM", i),
 							level=logging.ERROR, noiselevel=-1)
 					writemsg_level("\n", level=logging.ERROR, noiselevel=-1)
