@@ -1,9 +1,8 @@
 # portage.py -- core Portage functionality
-# Copyright 1998-2009 Gentoo Foundation
+# Copyright 1998-2010 Gentoo Foundation
 # Distributed under the terms of the GNU General Public License v2
-# $Id: __init__.py 15709 2010-03-02 21:13:50Z zmedico $
 
-VERSION="$Rev: 15709 $"[6:-2] + "-svn"
+VERSION="HEAD"
 
 # ===========================================================================
 # START OF IMPORTS -- START OF IMPORTS -- START OF IMPORTS -- START OF IMPORT
@@ -58,15 +57,10 @@ except ImportError as e:
 
 try:
 
-	try:
-		from collections import OrderedDict
-	except ImportError:
-		from portage.cache.mappings import OrderedDict
-
-	from portage.cache.cache_errors import CacheError
 	import portage.proxy.lazyimport
 	import portage.proxy as proxy
 	proxy.lazyimport.lazyimport(globals(),
+		'portage.cache.cache_errors:CacheError',
 		'portage.checksum',
 		'portage.checksum:perform_checksum,perform_md5,prelink_capable',
 		'portage.cvstree',
@@ -87,7 +81,6 @@ try:
 			'match_from_list,match_to_list',
 		'portage.dep.dep_check:dep_check,dep_eval,dep_wordreduce,dep_zapdeps',
 		'portage.eclass_cache',
-		'portage.env.loaders',
 		'portage.exception',
 		'portage.getbinpkg',
 		'portage.locks',
@@ -132,10 +125,15 @@ try:
 			'cpv_getkey@getCPFromCPV,endversion_keys,' + \
 			'suffix_value@endversion,pkgcmp,pkgsplit,vercmp,ververify',
 		'portage.xpak',
-		'portage._deprecated:commit_mtimedb,dep_virtual,' + \
-			'digestParseFile,getvirtuals,pkgmerge',
+		'time',
 	)
 
+	try:
+		from collections import OrderedDict
+	except ImportError:
+		proxy.lazyimport.lazyimport(globals(),
+			'portage.cache.mappings:OrderedDict')
+
 	import portage.const
 	from portage.const import VDB_PATH, PRIVATE_PATH, CACHE_PATH, DEPCACHE_PATH, \
 		USER_CONFIG_PATH, MODULES_FILE_PATH, CUSTOM_PROFILE_PATH, PORTAGE_BASE_PATH, \
@@ -227,7 +225,9 @@ class _unicode_func_wrapper(object):
 
 		rval = self._func(*wrapped_args, **wrapped_kwargs)
 
-		if isinstance(rval, (list, tuple)):
+		# Don't use isinstance() since we don't want to convert subclasses
+		# of tuple such as posix.stat_result in python-3.2.
+		if rval.__class__ in (list, tuple):
 			decoded_rval = []
 			for x in rval:
 				try:
@@ -290,6 +290,7 @@ class _unicode_module_wrapper(object):
 import os as _os
 _os_overrides = {
 	id(_os.fdopen)        : _os.fdopen,
+	id(_os.mkfifo)        : _os.mkfifo,
 	id(_os.popen)         : _os.popen,
 	id(_os.read)          : _os.read,
 	id(_os.system)        : _os.system,
@@ -325,6 +326,10 @@ except (ImportError, OSError) as e:
 # END OF IMPORTS -- END OF IMPORTS -- END OF IMPORTS -- END OF IMPORTS -- END
 # ===========================================================================
 
+_python_interpreter = os.path.realpath(sys.executable)
+_bin_path = PORTAGE_BIN_PATH
+_pym_path = PORTAGE_PYM_PATH
+
 def _ensure_default_encoding():
 
 	default_encoding = sys.getdefaultencoding().lower().replace('-', '_')
@@ -470,20 +475,6 @@ def _parse_eapi_ebuild_head(f):
 			break
 	return '0'
 
-# True when FEATURES=parse-eapi-glep-55 is enabled.
-_glep_55_enabled = False
-
-_split_ebuild_name_glep55_re = re.compile(r'^(.*)\.ebuild(-([^.]+))?$')
-
-def _split_ebuild_name_glep55(name):
-	"""
-	@returns: (pkg-ver-rev, eapi)
-	"""
-	m = _split_ebuild_name_glep55_re.match(name)
-	if m is None:
-		return (None, None)
-	return (m.group(1), m.group(3))
-
 def _movefile(src, dest, **kwargs):
 	"""Calls movefile and raises a PortageException if an error occurs."""
 	if movefile(src, dest, **kwargs) is None:
@@ -493,7 +484,7 @@ def _movefile(src, dest, **kwargs):
 auxdbkeys = (
   'DEPEND',    'RDEPEND',   'SLOT',      'SRC_URI',
 	'RESTRICT',  'HOMEPAGE',  'LICENSE',   'DESCRIPTION',
-	'KEYWORDS',  'INHERITED', 'IUSE', 'UNUSED_00',
+	'KEYWORDS',  'INHERITED', 'IUSE', 'REQUIRED_USE',
 	'PDEPEND',   'PROVIDE', 'EAPI',
 	'PROPERTIES', 'DEFINED_PHASES', 'UNUSED_05', 'UNUSED_04',
 	'UNUSED_03', 'UNUSED_02', 'UNUSED_01',
@@ -503,14 +494,6 @@ auxdbkeylen=len(auxdbkeys)
 def portageexit():
 	if data.secpass > 1 and os.environ.get("SANDBOX_ON") != "1":
 		close_portdbapi_caches()
-		try:
-			mtimedb
-		except NameError:
-			pass
-		else:
-			mtimedb.commit()
-
-atexit_register(portageexit)
 
 def create_trees(config_root=None, target_root=None, trees=None):
 	if trees is None:
@@ -534,7 +517,8 @@ def create_trees(config_root=None, target_root=None, trees=None):
 		# environment to apply to the config that's associated
 		# with ROOT != "/", so pass a nearly empty dict for the env parameter.
 		clean_env = {}
-		for k in ('PATH', 'TERM'):
+		for k in ('PATH', 'PORTAGE_GRPNAME', 'PORTAGE_USERNAME', 'TERM',
+			'ftp_proxy', 'http_proxy', 'no_proxy'):
 			v = settings.get(k)
 			if v is not None:
 				clean_env[k] = v
@@ -544,7 +528,7 @@ def create_trees(config_root=None, target_root=None, trees=None):
 
 	for myroot, mysettings in myroots:
 		trees[myroot] = portage.util.LazyItemsDict(trees.get(myroot, {}))
-		trees[myroot].addLazySingleton("virtuals", mysettings.getvirtuals, myroot)
+		trees[myroot].addLazySingleton("virtuals", mysettings.getvirtuals)
 		trees[myroot].addLazySingleton(
 			"vartree", vartree, myroot, categories=mysettings.categories,
 				settings=mysettings)
@@ -554,6 +538,46 @@ def create_trees(config_root=None, target_root=None, trees=None):
 			binarytree, myroot, mysettings["PKGDIR"], settings=mysettings)
 	return trees
 
+if VERSION == 'HEAD':
+	class _LazyVersion(proxy.objectproxy.ObjectProxy):
+		def _get_target(self):
+			global VERSION
+			if VERSION is not self:
+				return VERSION
+			if os.path.isdir(os.path.join(PORTAGE_BASE_PATH, '.git')):
+				status, output = subprocess_getstatusoutput((
+					"cd %s ; git describe --tags || exit $? ; " + \
+					"if [ -n \"`git diff-index --name-only --diff-filter=M HEAD`\" ] ; " + \
+					"then echo modified ; git rev-list --format=%%ct -n 1 HEAD ; fi ; " + \
+					"exit 0") % _shell_quote(PORTAGE_BASE_PATH))
+				if os.WIFEXITED(status) and os.WEXITSTATUS(status) == os.EX_OK:
+					output_lines = output.splitlines()
+					if output_lines:
+						version_split = output_lines[0].split('-')
+						if version_split:
+							VERSION = version_split[0].lstrip('v')
+							patchlevel = False
+							if len(version_split) > 1:
+								patchlevel = True
+								VERSION = "%s_p%s" %(VERSION, version_split[1])
+							if len(output_lines) > 1 and output_lines[1] == 'modified':
+								head_timestamp = None
+								if len(output_lines) > 3:
+									try:
+										head_timestamp = long(output_lines[3])
+									except ValueError:
+										pass
+								timestamp = long(time.time())
+								if head_timestamp is not None and timestamp > head_timestamp:
+									timestamp = timestamp - head_timestamp
+								if not patchlevel:
+									VERSION = "%s_p0" % (VERSION,)
+								VERSION = "%s_p%d" % (VERSION, timestamp)
+							return VERSION
+			VERSION = 'HEAD'
+			return VERSION
+	VERSION = _LazyVersion()
+
 class _LegacyGlobalProxy(proxy.objectproxy.ObjectProxy):
 
 	__slots__ = ('_name',)
@@ -570,7 +594,7 @@ class _LegacyGlobalProxy(proxy.objectproxy.ObjectProxy):
 _legacy_global_var_names = ("archlist", "db", "features",
 	"groups", "mtimedb", "mtimedbfile", "pkglines",
 	"portdb", "profiledir", "root", "selinux_enabled",
-	"settings", "thirdpartymirrors", "usedefaults")
+	"settings", "thirdpartymirrors")
 
 for k in _legacy_global_var_names:
 	globals()[k] = _LegacyGlobalProxy(k)
