@@ -1,6 +1,5 @@
-# Copyright 1999-2009 Gentoo Foundation
+# Copyright 1999-2010 Gentoo Foundation
 # Distributed under the terms of the GNU General Public License v2
-# $Id: _selinux.py 15109 2009-12-17 03:36:16Z zmedico $
 
 # Don't use the unicode-wrapped os and shutil modules here since
 # the whole _selinux module itself will be wrapped.
@@ -14,7 +13,6 @@ from portage import _unicode_encode
 from portage.localization import _
 
 import selinux
-from selinux import is_selinux_enabled
 
 def copyfile(src, dest):
 	src = _unicode_encode(src, encoding=_encodings['fs'], errors='strict')
@@ -37,6 +35,9 @@ def getcontext():
 
 	return ctx
 
+def is_selinux_enabled():
+	return selinux.is_selinux_enabled()
+
 def mkdir(target, refdir):
 	target = _unicode_encode(target, encoding=_encodings['fs'], errors='strict')
 	refdir = _unicode_encode(refdir, encoding=_encodings['fs'], errors='strict')
