@@ -1,41 +1,52 @@
-# Copyright 1998-2009 Gentoo Foundation
+# Copyright 1998-2010 Gentoo Foundation
 # Distributed under the terms of the GNU General Public License v2
-# $Id: bintree.py 15696 2010-03-02 21:09:35Z zmedico $
 
 __all__ = ["bindbapi", "binarytree"]
 
 import portage
 portage.proxy.lazyimport.lazyimport(globals(),
+	'portage.checksum:hashfunc_map,perform_multiple_checksums,verify_all',
 	'portage.dbapi.dep_expand:dep_expand',
 	'portage.dep:dep_getkey,isjustname,match_from_list',
 	'portage.output:EOutput,colorize',
+	'portage.locks:lockfile,unlockfile',
 	'portage.package.ebuild.doebuild:_vdb_use_conditional_atoms',
+	'portage.package.ebuild.fetch:_check_distfile',
 	'portage.update:update_dbentries',
-	'portage.util:ensure_dirs,normalize_path,writemsg,writemsg_stdout',
+	'portage.util:atomic_ofstream,ensure_dirs,normalize_path,' + \
+		'writemsg,writemsg_stdout',
 	'portage.util.listdir:listdir',
 	'portage.versions:best,catpkgsplit,catsplit',
 )
 
 from portage.cache.mappings import slot_dict_class
+from portage.const import CACHE_PATH
 from portage.dbapi.virtual import fakedbapi
+from portage.dep import Atom, use_reduce, paren_enclose
 from portage.exception import InvalidPackageName, \
 	PermissionDenied, PortageException
 from portage.localization import _
-
 from portage import _movefile
 from portage import os
 from portage import _encodings
 from portage import _unicode_decode
 from portage import _unicode_encode
-from portage.package.ebuild.fetch import _check_distfile
 
 import codecs
 import errno
 import re
 import stat
+import subprocess
 import sys
-import warnings
+import tempfile
+import textwrap
 from itertools import chain
+try:
+	from urllib.parse import urlparse
+	from urllib.request import urlopen as urllib_request_urlopen
+except ImportError:
+	from urlparse import urlparse
+	from urllib import urlopen as urllib_request_urlopen
 
 if sys.hexversion >= 0x3000000:
 	basestring = str
@@ -54,7 +65,8 @@ class bindbapi(fakedbapi):
 		self._aux_cache_keys = set(
 			["BUILD_TIME", "CHOST", "DEPEND", "EAPI", "IUSE", "KEYWORDS",
 			"LICENSE", "PDEPEND", "PROPERTIES", "PROVIDE",
-			"RDEPEND", "repository", "RESTRICT", "SLOT", "USE"])
+			"RDEPEND", "repository", "RESTRICT", "SLOT", "USE", "DEFINED_PHASES",
+			"REQUIRED_USE"])
 		self._aux_cache_slot_dict = slot_dict_class(self._aux_cache_keys)
 		self._aux_cache = {}
 
@@ -63,6 +75,11 @@ class bindbapi(fakedbapi):
 			self.bintree.populate()
 		return fakedbapi.match(self, *pargs, **kwargs)
 
+	def cpv_exists(self, cpv):
+		if self.bintree and not self.bintree.populated:
+			self.bintree.populate()
+		return fakedbapi.cpv_exists(self, cpv)
+
 	def cpv_inject(self, cpv, **kwargs):
 		self._aux_cache.pop(cpv, None)
 		fakedbapi.cpv_inject(self, cpv, **kwargs)
@@ -159,6 +176,41 @@ class bindbapi(fakedbapi):
 			self.bintree.populate()
 		return fakedbapi.cpv_all(self)
 
+def _pkgindex_cpv_map_latest_build(pkgindex):
+	"""
+	Given a PackageIndex instance, create a dict of cpv -> metadata map.
+	If multiple packages have identical CPV values, prefer the package
+	with latest BUILD_TIME value.
+	@param pkgindex: A PackageIndex instance.
+	@type pkgindex: PackageIndex
+	@rtype: dict
+	@returns: a dict containing entry for the give cpv.
+	"""
+	cpv_map = {}
+
+	for d in pkgindex.packages:
+		cpv = d["CPV"]
+
+		btime = d.get('BUILD_TIME', '')
+		try:
+			btime = int(btime)
+		except ValueError:
+			btime = None
+
+		other_d = cpv_map.get(cpv)
+		if other_d is not None:
+			other_btime = other_d.get('BUILD_TIME', '')
+			try:
+				other_btime = int(other_btime)
+			except ValueError:
+				other_btime = None
+			if other_btime and (not btime or other_btime > btime):
+				continue
+
+		cpv_map[cpv] = d
+
+	return cpv_map
+
 class binarytree(object):
 	"this tree scans for a list of all packages available in PKGDIR"
 	def __init__(self, root, pkgdir, virtual=None, settings=None):
@@ -189,7 +241,8 @@ class binarytree(object):
 			self._pkgindex_aux_keys = \
 				["BUILD_TIME", "CHOST", "DEPEND", "DESCRIPTION", "EAPI",
 				"IUSE", "KEYWORDS", "LICENSE", "PDEPEND", "PROPERTIES",
-				"PROVIDE", "RDEPEND", "repository", "SLOT", "USE"]
+				"PROVIDE", "RDEPEND", "repository", "SLOT", "USE", "DEFINED_PHASES",
+				"REQUIRED_USE"]
 			self._pkgindex_aux_keys = list(self._pkgindex_aux_keys)
 			self._pkgindex_use_evaluated_keys = \
 				("LICENSE", "RDEPEND", "DEPEND",
@@ -213,7 +266,9 @@ class binarytree(object):
 				"RDEPEND" : "",
 				"RESTRICT": "",
 				"SLOT"    : "0",
-				"USE"     : ""
+				"USE"     : "",
+				"DEFINED_PHASES" : "",
+				"REQUIRED_USE" : ""
 			}
 			self._pkgindex_inherited_keys = ["CHOST", "repository"]
 			self._pkgindex_default_header_data = {
@@ -234,25 +289,7 @@ class binarytree(object):
 				chain(*self._pkgindex_translated_keys)
 			))
 
-	def _get_remotepkgs(self):
-		warnings.warn("Use binarytree._remotepkgs insead of binarytree.remotepkgs",
-			DeprecationWarning)
-		return self.__remotepkgs
-
-	def _set_remotepkgs(self, remotepkgs):
-		warnings.warn("Use binarytree._remotepkgs insead of binarytree.remotepkgs",
-			DeprecationWarning)
-		self.__remotepkgs = remotepkgs
-
-	def _del_remotepkgs(self):
-		warnings.warn("Use binarytree._remotepkgs insead of binarytree.remotepkgs",
-			DeprecationWarning)
-		del self.__remotepkgs
-
-	remotepkgs = property(_get_remotepkgs, _set_remotepkgs, _del_remotepkgs,
-		"Deprecated self.remotepkgs, only for backward compatibility")
-
-	def move_ent(self, mylist):
+	def move_ent(self, mylist, repo_match=None):
 		if not self.populated:
 			self.populate()
 		origcp = mylist[1]
@@ -271,6 +308,10 @@ class binarytree(object):
 			if mycpv_cp != origcp:
 				# Ignore PROVIDE virtual match.
 				continue
+			if repo_match is not None \
+				and not repo_match(self.dbapi.aux_get(mycpv,
+					['repository'])[0]):
+				continue
 			mynewcpv = mycpv.replace(mycpv_cp, str(newcp), 1)
 			myoldpkg = catsplit(mycpv)[1]
 			mynewpkg = catsplit(mynewcpv)[1]
@@ -461,18 +502,12 @@ class binarytree(object):
 		_movefile(src_path, dest_path, mysettings=self.settings)
 		self._pkg_paths[cpv] = mypath
 
-	def populate(self, getbinpkgs=0, getbinpkgsonly=None):
+	def populate(self, getbinpkgs=0):
 		"populates the binarytree"
 
-		if getbinpkgsonly is not None:
-			warnings.warn(
-				"portage.dbapi.bintree.binarytree.populate(): " + \
-				"getbinpkgsonly parameter is deprecated",
-				DeprecationWarning, stacklevel=2)
-
 		if self._populating:
 			return
-		from portage.locks import lockfile, unlockfile
+
 		pkgindex_lock = None
 		try:
 			if os.access(self.pkgdir, os.W_OK):
@@ -617,8 +652,7 @@ class binarytree(object):
 								", ".join(missing_keys))
 						msg.append(_(" This binary package is not " \
 							"recoverable and should be deleted."))
-						from textwrap import wrap
-						for line in wrap("".join(msg), 72):
+						for line in textwrap.wrap("".join(msg), 72):
 							writemsg("!!! %s\n" % line, noiselevel=-1)
 						self.invalids.append(mypkg)
 						continue
@@ -696,12 +730,9 @@ class binarytree(object):
 				del pkgindex.packages[:]
 				pkgindex.packages.extend(iter(metadata.values()))
 				self._update_pkgindex_header(pkgindex.header)
-				from portage.util import atomic_ofstream
 				f = atomic_ofstream(self._pkgindex_file)
-				try:
-					pkgindex.write(f)
-				finally:
-					f.close()
+				pkgindex.write(f)
+				f.close()
 
 		if getbinpkgs and not self.settings["PORTAGE_BINHOST"]:
 			writemsg(_("!!! PORTAGE_BINHOST unset, but use is requested.\n"),
@@ -709,14 +740,16 @@ class binarytree(object):
 
 		if getbinpkgs and 'PORTAGE_BINHOST' in self.settings:
 			base_url = self.settings["PORTAGE_BINHOST"]
-			from portage.const import CACHE_PATH
-			try:
-				from urllib.parse import urlparse
-			except ImportError:
-				from urlparse import urlparse
-			urldata = urlparse(base_url)
-			pkgindex_file = os.path.join(self.settings["ROOT"], CACHE_PATH, "binhost",
-				urldata[1] + urldata[2], "Packages")
+			parsed_url = urlparse(base_url)
+			host = parsed_url.netloc
+			port = parsed_url.port
+			port_args = []
+			if port is not None:
+				port_str = ":%s" % (port,)
+				if host.endswith(port_str):
+					host = host[:-len(port_str)]
+			pkgindex_file = os.path.join(self.settings["EROOT"], CACHE_PATH, "binhost",
+				host, parsed_url.path.lstrip("/"), "Packages")
 			pkgindex = self._new_pkgindex()
 			try:
 				f = codecs.open(_unicode_encode(pkgindex_file,
@@ -731,16 +764,40 @@ class binarytree(object):
 				if e.errno != errno.ENOENT:
 					raise
 			local_timestamp = pkgindex.header.get("TIMESTAMP", None)
-			try:
-				from urllib.request import urlopen as urllib_request_urlopen
-			except ImportError:
-				from urllib import urlopen as urllib_request_urlopen
 			rmt_idx = self._new_pkgindex()
+			proc = None
+			tmp_filename = None
 			try:
 				# urlparse.urljoin() only works correctly with recognized
 				# protocols and requires the base url to have a trailing
 				# slash, so join manually...
-				f = urllib_request_urlopen(base_url.rstrip("/") + "/Packages")
+				try:
+					f = urllib_request_urlopen(base_url.rstrip("/") + "/Packages")
+				except IOError:
+					path = parsed_url.path.rstrip("/") + "/Packages"
+
+					if parsed_url.scheme == 'sftp':
+						# The sftp command complains about 'Illegal seek' if
+						# we try to make it write to /dev/stdout, so use a
+						# temp file instead.
+						fd, tmp_filename = tempfile.mkstemp()
+						os.close(fd)
+						if port is not None:
+							port_args = ['-P', "%s" % (port,)]
+						proc = subprocess.Popen(['sftp'] + port_args + \
+							[host + ":" + path, tmp_filename])
+						if proc.wait() != os.EX_OK:
+							raise
+						f = open(tmp_filename, 'rb')
+					elif parsed_url.scheme == 'ssh':
+						if port is not None:
+							port_args = ['-p', "%s" % (port,)]
+						proc = subprocess.Popen(['ssh'] + port_args + \
+							[host, '--', 'cat', path], stdout=subprocess.PIPE)
+						f = proc.stdout
+					else:
+						raise
+
 				f_dec = codecs.iterdecode(f,
 					_encodings['repo.content'], errors='replace')
 				try:
@@ -766,24 +823,31 @@ class binarytree(object):
 				writemsg("!!! %s\n\n" % str(e))
 				del e
 				pkgindex = None
+			if proc is not None:
+				if proc.poll() is None:
+					proc.kill()
+					proc.wait()
+				proc = None
+			if tmp_filename is not None:
+				try:
+					os.unlink(tmp_filename)
+				except OSError:
+					pass
 			if pkgindex is rmt_idx:
 				pkgindex.modified = False # don't update the header
-				from portage.util import atomic_ofstream, ensure_dirs
 				try:
 					ensure_dirs(os.path.dirname(pkgindex_file))
 					f = atomic_ofstream(pkgindex_file)
 					pkgindex.write(f)
 					f.close()
-				except PortageException:
-					if os.access(os.path.join(
-						self.settings["ROOT"], CACHE_PATH), os.W_OK):
+				except (IOError, PortageException):
+					if os.access(os.path.dirname(pkgindex_file), os.W_OK):
 						raise
 					# The current user doesn't have permission to cache the
 					# file, but that's alright.
 			if pkgindex:
-				self._remotepkgs = {}
-				for d in pkgindex.packages:
-					self._remotepkgs[d["CPV"]] = d
+				# Organize remote package list as a cpv -> metadata map.
+				self._remotepkgs = _pkgindex_cpv_map_latest_build(pkgindex)
 				self._remote_has_index = True
 				self._remote_base_uri = pkgindex.header.get("URI", base_url)
 				self.__remotepkgs = {}
@@ -925,7 +989,6 @@ class binarytree(object):
 
 		# Reread the Packages index (in case it's been changed by another
 		# process) and then updated it, all while holding a lock.
-		from portage.locks import lockfile, unlockfile
 		pkgindex_lock = None
 		created_symlink = False
 		try:
@@ -933,8 +996,13 @@ class binarytree(object):
 				wantnewlockfile=1)
 			if filename is not None:
 				new_filename = self.getname(cpv)
-				self._ensure_dir(os.path.dirname(new_filename))
-				_movefile(filename, new_filename, mysettings=self.settings)
+				try:
+					samefile = os.path.samefile(filename, new_filename)
+				except OSError:
+					samefile = False
+				if not samefile:
+					self._ensure_dir(os.path.dirname(new_filename))
+					_movefile(filename, new_filename, mysettings=self.settings)
 			if self._all_directory and \
 				self.getname(cpv).split(os.path.sep)[-2] == "All":
 				self._create_symlink(cpv)
@@ -944,6 +1012,16 @@ class binarytree(object):
 			if not self._pkgindex_version_supported(pkgindex):
 				pkgindex = self._new_pkgindex()
 
+			# Discard remote metadata to ensure that _pkgindex_entry
+			# gets the local metadata. This also updates state for future
+			# isremote calls.
+			if self._remotepkgs is not None:
+				self._remotepkgs.pop(cpv, None)
+
+			# Discard cached metadata to ensure that _pkgindex_entry
+			# doesn't return stale metadata.
+			self.dbapi._aux_cache.pop(cpv, None)
+
 			try:
 				d = self._pkgindex_entry(cpv)
 			except portage.exception.InvalidDependString:
@@ -972,21 +1050,13 @@ class binarytree(object):
 			pkgindex.packages.append(d)
 
 			self._update_pkgindex_header(pkgindex.header)
-			from portage.util import atomic_ofstream
 			f = atomic_ofstream(os.path.join(self.pkgdir, "Packages"))
-			try:
-				pkgindex.write(f)
-			finally:
-				f.close()
+			pkgindex.write(f)
+			f.close()
 		finally:
 			if pkgindex_lock:
 				unlockfile(pkgindex_lock)
 
-		if self._remotepkgs is not None:
-			# When a remote package is downloaded and injected,
-			# update state so self.isremote() returns False.
-			self._remotepkgs.pop(cpv, None)
-
 	def _pkgindex_entry(self, cpv):
 		"""
 		Performs checksums and evaluates USE flag conditionals.
@@ -996,7 +1066,6 @@ class binarytree(object):
 		"""
 
 		pkg_path = self.getname(cpv)
-		from portage.checksum import perform_multiple_checksums
 
 		d = dict(zip(self._pkgindex_aux_keys,
 			self.dbapi.aux_get(cpv, self._pkgindex_aux_keys)))
@@ -1064,28 +1133,20 @@ class binarytree(object):
 		use = [f for f in use if f in iuse]
 		use.sort()
 		metadata["USE"] = " ".join(use)
-		from portage.dep import paren_reduce, use_reduce, \
-			paren_normalize, paren_enclose
 		for k in self._pkgindex_use_evaluated_keys:
+			if k.endswith('DEPEND'):
+				token_class = Atom
+			else:
+				token_class = None
+
 			try:
-				deps = paren_reduce(metadata[k])
-				deps = use_reduce(deps, uselist=raw_use)
-				deps = paren_normalize(deps)
+				deps = metadata[k]
+				deps = use_reduce(deps, uselist=raw_use, token_class=token_class)
 				deps = paren_enclose(deps)
 			except portage.exception.InvalidDependString as e:
 				writemsg("%s: %s\n" % (k, str(e)),
 					noiselevel=-1)
 				raise
-			if k in _vdb_use_conditional_atoms:
-				v_split = []
-				for x in deps.split():
-					try:
-						x = portage.dep.Atom(x)
-					except portage.exception.InvalidAtom:
-						v_split.append(x)
-					else:
-						v_split.append(str(x.evaluate_conditionals(raw_use)))
-				deps = ' '.join(v_split)
 			metadata[k] = deps
 
 	def exists_specific(self, cpv):
@@ -1157,10 +1218,6 @@ class binarytree(object):
 		
 		mydest = os.path.dirname(self.getname(pkgname))
 		self._ensure_dir(mydest)
-		try:
-			from urllib.parse import urlparse
-		except ImportError:
-			from urlparse import urlparse
 		# urljoin doesn't work correctly with unrecognized protocols like sftp
 		if self._remote_has_index:
 			rel_url = self._remotepkgs[pkgname].get("PATH")
@@ -1226,7 +1283,6 @@ class binarytree(object):
 			return False
 
 		digests = {}
-		from portage.checksum import hashfunc_map, verify_all
 		for k in hashfunc_map:
 			v = metadata.get(k)
 			if not v:
