@@ -1,6 +1,5 @@
-# Copyright 2004-2009 Gentoo Foundation
+# Copyright 2004-2010 Gentoo Foundation
 # Distributed under the terms of the GNU General Public License v2
-# $Id: __init__.py 15719 2010-03-02 21:38:01Z zmedico $
 
 __all__ = ['apply_permissions', 'apply_recursive_permissions',
 	'apply_secpass_permissions', 'apply_stat_permissions', 'atomic_ofstream',
@@ -9,12 +8,16 @@ __all__ = ['apply_permissions', 'apply_recursive_permissions',
 	'grabdict_package', 'grabfile', 'grabfile_package', 'grablines',
 	'initialize_logger', 'LazyItemsDict', 'map_dictlist_vals',
 	'new_protect_filename', 'normalize_path', 'pickle_read', 'stack_dictlist',
-	'stack_dicts', 'stack_lists', 'unique_array', 'varexpand', 'write_atomic',
-	'writedict', 'writemsg', 'writemsg_level', 'writemsg_stdout']
+	'stack_dicts', 'stack_lists', 'unique_array', 'unique_everseen', 'varexpand',
+	'write_atomic', 'writedict', 'writemsg', 'writemsg_level', 'writemsg_stdout']
 
 import codecs
 from copy import deepcopy
 import errno
+try:
+	from itertools import filterfalse
+except ImportError:
+	from itertools import ifilterfalse as filterfalse
 import logging
 import re
 import shlex
@@ -24,6 +27,10 @@ import sys
 import traceback
 
 import portage
+portage.proxy.lazyimport.lazyimport(globals(),
+	'portage.dep:Atom',
+	'portage.util.listdir:_ignorecvs_dirs'
+)
 from portage import StringIO
 from portage import os
 from portage import pickle
@@ -34,7 +41,6 @@ from portage import _unicode_encode
 from portage import _unicode_decode
 from portage.exception import InvalidAtom, PortageException, FileNotFound, \
        OperationNotPermitted, PermissionDenied, ReadOnlyFileSystem
-from portage.dep import Atom
 from portage.localization import _
 from portage.proxy.objectproxy import ObjectProxy
 from portage.cache.mappings import UserDict
@@ -57,10 +63,14 @@ def writemsg(mystr,noiselevel=0,fd=None):
 		fd = sys.stderr
 	if noiselevel <= noiselimit:
 		# avoid potential UnicodeEncodeError
-		mystr = _unicode_encode(mystr,
-			encoding=_encodings['stdio'], errors='backslashreplace')
-		if sys.hexversion >= 0x3000000:
-			fd = fd.buffer
+		if isinstance(fd, StringIO):
+			mystr = _unicode_decode(mystr,
+				encoding=_encodings['content'], errors='replace')
+		else:
+			mystr = _unicode_encode(mystr,
+				encoding=_encodings['stdio'], errors='backslashreplace')
+			if sys.hexversion >= 0x3000000 and fd in (sys.stdout, sys.stderr):
+				fd = fd.buffer
 		fd.write(mystr)
 		fd.flush()
 
@@ -105,13 +115,14 @@ def normalize_path(mypath):
 	else:
 		return os.path.normpath(mypath)
 
-def grabfile(myfilename, compat_level=0, recursive=0):
+def grabfile(myfilename, compat_level=0, recursive=0, remember_source_file=False):
 	"""This function grabs the lines in a file, normalizes whitespace and returns lines in a list; if a line
 	begins with a #, it is ignored, as are empty lines"""
 
-	mylines=grablines(myfilename, recursive)
+	mylines=grablines(myfilename, recursive, remember_source_file=True)
 	newlines=[]
-	for x in mylines:
+
+	for x, source_file in mylines:
 		#the split/join thing removes leading and trailing whitespace, and converts any whitespace in the line
 		#into single spaces.
 		myline = _unicode_decode(' ').join(x.split())
@@ -131,7 +142,10 @@ def grabfile(myfilename, compat_level=0, recursive=0):
 				continue
 			else:
 				continue
-		newlines.append(myline)
+		if remember_source_file:
+			newlines.append((myline, source_file))
+		else:
+			newlines.append(myline)
 	return newlines
 
 def map_dictlist_vals(func,myDict):
@@ -217,25 +231,59 @@ def stack_dicts(dicts, incremental=0, incrementals=[], ignore_none=0):
 				final_dict[k]  = v
 	return final_dict
 
-def stack_lists(lists, incremental=1):
+def stack_lists(lists, incremental=1, remember_source_file=False,
+	warn_for_unmatched_removal=False, strict_warn_for_unmatched_removal=False):
 	"""Stacks an array of list-types into one array. Optionally removing
 	distinct values using '-value' notation. Higher index is preferenced.
 
 	all elements must be hashable."""
-
+	matched_removals = set()
+	unmatched_removals = {}
 	new_list = {}
-	for x in lists:
-		for y in filter(None, x):
-			if incremental:
-				if y == "-*":
-					new_list.clear()
-				elif y[:1] == '-':
-					new_list.pop(y[1:], None)
-				else:
-					new_list[y] = True
+	for sub_list in lists:
+		for token in sub_list:
+			token_key = token
+			if remember_source_file:
+				token, source_file = token
 			else:
-				new_list[y] = True
-	return list(new_list)
+				source_file = False
+
+			if token is None:
+				continue
+
+			if incremental:
+				if token == "-*":
+					new_list.clear()
+				elif token[:1] == '-':
+					try:
+						new_list.pop(token[1:])
+					except KeyError:
+						if source_file and \
+							(strict_warn_for_unmatched_removal or \
+							token_key not in matched_removals):
+							unmatched_removals.setdefault(source_file, set()).add(token)
+					else:
+						matched_removals.add(token_key)
+				else:
+					new_list[token] = source_file
+			else:
+				new_list[token] = source_file
+
+	if warn_for_unmatched_removal:
+		for source_file, tokens in unmatched_removals.items():
+			if len(tokens) > 3:
+				selected = [tokens.pop(), tokens.pop(), tokens.pop()]
+				writemsg(_("--- Unmatch removal atoms in %s: %s and %s more\n") % \
+					(source_file, ", ".join(selected), len(tokens)),
+					noiselevel=-1)
+			else:
+				writemsg(_("--- Unmatch removal atom(s) in %s: %s\n") % (source_file, ", ".join(tokens)),
+					noiselevel=-1)
+
+	if remember_source_file:
+		return list(new_list.items())
+	else:
+		return list(new_list)
 
 def grabdict(myfilename, juststrings=0, empty=0, recursive=0, incremental=1):
 	"""
@@ -279,29 +327,63 @@ def grabdict(myfilename, juststrings=0, empty=0, recursive=0, incremental=1):
 			newdict[k] = " ".join(v)
 	return newdict
 
-def grabdict_package(myfilename, juststrings=0, recursive=0):
+def read_corresponding_eapi_file(filename):
+	"""
+	Read the 'eapi' file from the directory 'filename' is in.
+	Returns "0" if the file is not present or invalid.
+	"""
+	default = "0"
+	eapi_file = os.path.join(os.path.dirname(filename), "eapi")
+	try:
+		f = open(eapi_file, "r")
+		lines = f.readlines()
+		if len(lines) == 1:
+			eapi = lines[0]
+		else:
+			writemsg(_("--- Invalid 'eapi' file (doesn't contain exactly one line): %s\n") % (eapi_file),
+				noiselevel=-1)
+			eapi = default
+		f.close()
+	except IOError:
+		eapi = default
+
+	return eapi
+
+def grabdict_package(myfilename, juststrings=0, recursive=0, allow_wildcard=False, allow_repo=False,
+	verify_eapi=False, eapi=None):
 	""" Does the same thing as grabdict except it validates keys
 	    with isvalidatom()"""
 	pkgs=grabdict(myfilename, juststrings, empty=1, recursive=recursive)
+	if not pkgs:
+		return pkgs
+	if verify_eapi and eapi is None:
+		eapi = read_corresponding_eapi_file(myfilename)
+
 	# We need to call keys() here in order to avoid the possibility of
 	# "RuntimeError: dictionary changed size during iteration"
 	# when an invalid atom is deleted.
 	atoms = {}
 	for k, v in pkgs.items():
 		try:
-			k = Atom(k)
-		except InvalidAtom:
-			writemsg(_("--- Invalid atom in %s: %s\n") % (myfilename, k),
+			k = Atom(k, allow_wildcard=allow_wildcard, allow_repo=allow_repo, eapi=eapi)
+		except InvalidAtom as e:
+			writemsg(_("--- Invalid atom in %s: %s\n") % (myfilename, e),
 				noiselevel=-1)
 		else:
 			atoms[k] = v
 	return atoms
 
-def grabfile_package(myfilename, compatlevel=0, recursive=0):
-	pkgs=grabfile(myfilename, compatlevel, recursive=recursive)
+def grabfile_package(myfilename, compatlevel=0, recursive=0, allow_wildcard=False, allow_repo=False,
+	remember_source_file=False, verify_eapi=False, eapi=None):
+
+	pkgs=grabfile(myfilename, compatlevel, recursive=recursive, remember_source_file=True)
+	if not pkgs:
+		return pkgs
+	if verify_eapi and eapi is None:
+		eapi = read_corresponding_eapi_file(myfilename)
 	mybasename = os.path.basename(myfilename)
 	atoms = []
-	for pkg in pkgs:
+	for pkg, source_file in pkgs:
 		pkg_orig = pkg
 		# for packages and package.mask files
 		if pkg[:1] == "-":
@@ -309,36 +391,45 @@ def grabfile_package(myfilename, compatlevel=0, recursive=0):
 		if pkg[:1] == '*' and mybasename == 'packages':
 			pkg = pkg[1:]
 		try:
-			pkg = Atom(pkg)
-		except InvalidAtom:
-			writemsg(_("--- Invalid atom in %s: %s\n") % (myfilename, pkg),
+			pkg = Atom(pkg, allow_wildcard=allow_wildcard, allow_repo=allow_repo, eapi=eapi)
+		except InvalidAtom as e:
+			writemsg(_("--- Invalid atom in %s: %s\n") % (myfilename, e),
 				noiselevel=-1)
 		else:
 			if pkg_orig == str(pkg):
 				# normal atom, so return as Atom instance
-				atoms.append(pkg)
+				if remember_source_file:
+					atoms.append((pkg, source_file))
+				else:
+					atoms.append(pkg)
 			else:
 				# atom has special prefix, so return as string
-				atoms.append(pkg_orig)
+				if remember_source_file:
+					atoms.append((pkg_orig, source_file))
+				else:
+					atoms.append(pkg_orig)
 	return atoms
 
-def grablines(myfilename,recursive=0):
+def grablines(myfilename, recursive=0, remember_source_file=False):
 	mylines=[]
 	if recursive and os.path.isdir(myfilename):
-		if myfilename in ["RCS", "CVS", "SCCS"]:
+		if os.path.basename(myfilename) in _ignorecvs_dirs:
 			return mylines
 		dirlist = os.listdir(myfilename)
 		dirlist.sort()
 		for f in dirlist:
 			if not f.startswith(".") and not f.endswith("~"):
 				mylines.extend(grablines(
-					os.path.join(myfilename, f), recursive))
+					os.path.join(myfilename, f), recursive, remember_source_file))
 	else:
 		try:
 			myfile = codecs.open(_unicode_encode(myfilename,
 				encoding=_encodings['fs'], errors='strict'),
 				mode='r', encoding=_encodings['content'], errors='replace')
-			mylines = myfile.readlines()
+			if remember_source_file:
+				mylines = [(line, myfilename) for line in myfile.readlines()]
+			else:
+				mylines = myfile.readlines()
 			myfile.close()
 		except IOError as e:
 			if e.errno == PermissionDenied.errno:
@@ -399,9 +490,6 @@ def getconfig(mycfg, tolerant=0, allow_sourcing=False, expand=True):
 		expand_map = {}
 	mykeys = {}
 	try:
-		# Workaround for avoiding a silent error in shlex that
-		# is triggered by a source statement at the end of the file without a
-		# trailing newline after the source statement
 		# NOTE: shex doesn't seem to support unicode objects
 		# (produces spurious \0 characters with python-2.6.2)
 		if sys.hexversion < 0x3000000:
@@ -411,15 +499,27 @@ def getconfig(mycfg, tolerant=0, allow_sourcing=False, expand=True):
 			content = open(_unicode_encode(mycfg,
 				encoding=_encodings['fs'], errors='strict'), mode='r',
 				encoding=_encodings['content'], errors='replace').read()
-		if content and content[-1] != '\n':
-			content += '\n'
 	except IOError as e:
 		if e.errno == PermissionDenied.errno:
 			raise PermissionDenied(mycfg)
 		if e.errno != errno.ENOENT:
 			writemsg("open('%s', 'r'): %s\n" % (mycfg, e), noiselevel=-1)
-			raise
+			if e.errno not in (errno.EISDIR,):
+				raise
 		return None
+
+	# Workaround for avoiding a silent error in shlex that is
+	# triggered by a source statement at the end of the file
+	# without a trailing newline after the source statement.
+	if content and content[-1] != '\n':
+		content += '\n'
+
+	# Warn about dos-style line endings since that prevents
+	# people from being able to source them with bash.
+	if '\r' in content:
+		writemsg(("!!! " + _("Please use dos2unix to convert line endings " + \
+			"in config file: '%s'") + "\n") % mycfg, noiselevel=-1)
+
 	try:
 		if tolerant:
 			shlex_class = _tolerant_shlex
@@ -495,7 +595,9 @@ def getconfig(mycfg, tolerant=0, allow_sourcing=False, expand=True):
 	
 #cache expansions of constant strings
 cexpand={}
-def varexpand(mystring, mydict={}):
+def varexpand(mystring, mydict=None):
+	if mydict is None:
+		mydict = {}
 	newstring = cexpand.get(" "+mystring, None)
 	if newstring is not None:
 		return newstring
@@ -557,8 +659,14 @@ def varexpand(mystring, mydict={}):
 						newstring=newstring+chr(0o11)
 					elif a=='v':
 						newstring=newstring+chr(0o13)
+					elif a in ('\'', '"'):
+						# Quote removal is handled by shlex.
+						newstring = newstring + mystring[pos-2:pos]
+						continue
 					elif a!='\n':
-						#remove backslash only, as bash does: this takes care of \\ and \' and \" as well
+						# Remove backslash only, as bash does. This takes care
+						# of \\. Note that we don't handle quotes here since
+						# quote removal is handled by shlex.
 						newstring=newstring+mystring[pos-1:pos]
 						continue
 			elif (mystring[pos]=="$") and (mystring[pos-1]!="\\"):
@@ -709,6 +817,26 @@ def unique_array(s):
 			u.append(x)
 	return u
 
+def unique_everseen(iterable, key=None):
+    """
+    List unique elements, preserving order. Remember all elements ever seen.
+    Taken from itertools documentation.
+    """
+    # unique_everseen('AAAABBBCCDAABBB') --> A B C D
+    # unique_everseen('ABBCcAD', str.lower) --> A B C D
+    seen = set()
+    seen_add = seen.add
+    if key is None:
+        for element in filterfalse(seen.__contains__, iterable):
+            seen_add(element)
+            yield element
+    else:
+        for element in iterable:
+            k = key(element)
+            if k not in seen:
+                seen_add(k)
+                yield element
+
 def apply_permissions(filename, uid=-1, gid=-1, mode=-1, mask=-1,
 	stat_cached=None, follow_links=True):
 	"""Apply user, group, and mode bits to a file if the existing bits do not
@@ -1033,10 +1161,15 @@ def write_atomic(file_path, content, **kwargs):
 		else:
 			raise
 
-def ensure_dirs(dir_path, *args, **kwargs):
+def ensure_dirs(dir_path, **kwargs):
 	"""Create a directory and call apply_permissions.
 	Returns True if a directory is created or the permissions needed to be
-	modified, and False otherwise."""
+	modified, and False otherwise.
+
+	This function's handling of EEXIST errors makes it useful for atomic
+	directory creation, in which multiple processes may be competing to
+	create the same directory.
+	"""
 
 	created_dir = False
 
@@ -1045,17 +1178,27 @@ def ensure_dirs(dir_path, *args, **kwargs):
 		created_dir = True
 	except OSError as oe:
 		func_call = "makedirs('%s')" % dir_path
-		if oe.errno in (errno.EEXIST, errno.EISDIR):
+		if oe.errno in (errno.EEXIST,):
 			pass
-		elif oe.errno == errno.EPERM:
-			raise OperationNotPermitted(func_call)
-		elif oe.errno == errno.EACCES:
-			raise PermissionDenied(func_call)
-		elif oe.errno == errno.EROFS:
-			raise ReadOnlyFileSystem(func_call)
 		else:
-			raise
-	perms_modified = apply_permissions(dir_path, *args, **kwargs)
+			if os.path.isdir(dir_path):
+				# NOTE: DragonFly raises EPERM for makedir('/')
+				# and that is supposed to be ignored here.
+				# Also, sometimes mkdir raises EISDIR on FreeBSD
+				# and we want to ignore that too (bug #187518).
+				pass
+			elif oe.errno == errno.EPERM:
+				raise OperationNotPermitted(func_call)
+			elif oe.errno == errno.EACCES:
+				raise PermissionDenied(func_call)
+			elif oe.errno == errno.EROFS:
+				raise ReadOnlyFileSystem(func_call)
+			else:
+				raise
+	if kwargs:
+		perms_modified = apply_permissions(dir_path, **kwargs)
+	else:
+		perms_modified = False
 	return created_dir or perms_modified
 
 class LazyItemsDict(UserDict):
@@ -1148,23 +1291,10 @@ class LazyItemsDict(UserDict):
 
 	def __deepcopy__(self, memo=None):
 		"""
-		WARNING: If any of the lazy items contains a bound method then it's
-		typical for deepcopy() to raise an exception like this:
-
-			File "/usr/lib/python2.5/copy.py", line 189, in deepcopy
-				y = _reconstruct(x, rv, 1, memo)
-			File "/usr/lib/python2.5/copy.py", line 322, in _reconstruct
-				y = callable(*args)
-			File "/usr/lib/python2.5/copy_reg.py", line 92, in __newobj__
-				return cls.__new__(cls, *args)
-			TypeError: instancemethod expected at least 2 arguments, got 0
-
-		If deepcopy() needs to work, this problem can be avoided by
-		implementing lazy items with normal (non-bound) functions.
-
-		If deepcopy() raises a TypeError for a lazy item that has been added
-		via a call to addLazySingleton(), the singleton will be automatically
-		evaluated and deepcopy() will instead be called on the result.
+		This forces evaluation of each contained lazy item, and deepcopy of
+		the result. A TypeError is raised if any contained lazy item is not
+		a singleton, since it is not necessarily possible for the behavior
+		of this type of item to be safely preserved.
 		"""
 		if memo is None:
 			memo = {}
@@ -1172,19 +1302,13 @@ class LazyItemsDict(UserDict):
 		memo[id(self)] = result
 		for k in self:
 			k_copy = deepcopy(k, memo)
-			if k in self.lazy_items:
-				lazy_item = self.lazy_items[k]
-				try:
-					result.lazy_items[k_copy] = deepcopy(lazy_item, memo)
-				except TypeError:
-					if not lazy_item.singleton:
-						raise
-					UserDict.__setitem__(result,
-						k_copy, deepcopy(self[k], memo))
-				else:
-					UserDict.__setitem__(result, k_copy, None)
-			else:
-				UserDict.__setitem__(result, k_copy, deepcopy(self[k], memo))
+			lazy_item = self.lazy_items.get(k)
+			if lazy_item is not None:
+				if not lazy_item.singleton:
+					raise TypeError(_unicode_decode("LazyItemsDict " + \
+						"deepcopy is unsafe with lazy items that are " + \
+						"not singletons: key=%s value=%s") % (k, lazy_item,))
+			UserDict.__setitem__(result, k_copy, deepcopy(self[k], memo))
 		return result
 
 	class _LazyItem(object):
